# coding: utf-8
import random
from collections import defaultdict
from threading import Lock

import numpy as np

from supervisely import logger
from supervisely.annotation.annotation import Annotation
from supervisely.imaging import image as sly_image


def samples_by_tags(required_tags, project):
    """
    Split samples from project by tags
    :param required_tags: list of tags names
    :param project: supervisely `Project` class object
    :returns:
    """
    img_annotations_groups = defaultdict(list)
    for dataset in project:
        for item_name in dataset:
            item_paths = dataset.get_item_paths(item_name)
            ann = Annotation.load_json_file(path=item_paths.ann_path, project_meta=project.meta)
            img_tags = ann.img_tags
            for required_tag in required_tags:
                if img_tags.has_key(required_tag):
                    # TODO migrate to ItemPath objects for img_annotations_groups
                    img_annotations_groups[required_tag].append((item_paths.img_path, item_paths.ann_path))
    return img_annotations_groups


def ensure_samples_nonempty(samples, tag_name, project_meta):
    """

    :param samples: list of pairs (image path, annotation path).
    :type samples: list
    :param tag_name: tag name for messages.
    :type tag_name: str
    :param project_meta: input project meta object.
    :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
    :returns: None
    :rtype: None
    """
    if len(samples) < 1:
        raise RuntimeError('There are no annotations with tag "{}"'.format(tag_name))

    for _, ann_path in samples:
        ann = Annotation.load_json_file(ann_path, project_meta)
        if len(ann.labels) > 0:
            return

    raise RuntimeError('There are no objects in annotations with tag "{}"'.format(tag_name))


class CorruptedSampleCatcher(object):
    """Context helper that tolerates up to N corrupted samples before raising; tracks failed UIDs."""

    def __init__(self, allow_corrupted_cnt):
        """
        :param allow_corrupted_cnt: Max allowed corrupted samples before raising.
        :type allow_corrupted_cnt: int
        """
        self.fails_allowed = allow_corrupted_cnt
        self._failed_uids = set()
        self._lock = Lock()

    def exec(self, uid, log_dct, f, *args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            self._lock.acquire()
            if uid not in self._failed_uids:
                self._failed_uids.add(uid)
                logger.warn('Sample processing error.', exc_info=True, extra={**log_dct, 'exc_str': str(e)})
            fail_cnt = len(self._failed_uids)
            self._lock.release()

            if fail_cnt > self.fails_allowed:
                raise RuntimeError('Too many errors occurred while processing samples. '
                                   'Allowed: {}.'.format(self.fails_allowed))


class SlyDataset:
    """
    SlyDataset class is mostly compatible with PyTorch 'Dataset' class except the output image format is HWC, not CHW.

    Override _get_sample_impl to convert raw Numpy arrays to the proper tensor format for your model.
    """

    def __init__(self, project_meta, samples, out_size, class_mapping, bkg_color, allow_corrupted_cnt=0,
                 catcher_retries=100):
        """
        :param project_meta: Project meta.
        :type project_meta: ProjectMeta
        :param samples: List of (img_path, ann_path).
        :type samples: list
        :param out_size: Output size (H, W).
        :type out_size: tuple
        :param class_mapping: Class name mapping.
        :type class_mapping: dict
        :param bkg_color: Background color (-1 for no background).
        :type bkg_color: int
        :param allow_corrupted_cnt: Max corrupted samples.
        :type allow_corrupted_cnt: int
        :param catcher_retries: Retries for catcher.
        :type catcher_retries: int
        """
        self._project_meta = project_meta
        self._samples = samples
        self._out_size = tuple(out_size)
        self._class_mapping = class_mapping
        self._bkg_color = bkg_color
        self._sample_catcher = CorruptedSampleCatcher(allow_corrupted_cnt)
        self._catcher_retries = catcher_retries

    def __len__(self):
        return len(self._samples)

    def load_annotation(self, fpath):
        # will not resize figures: resize gt instead
        return Annotation.load_json_file(fpath, self._project_meta)

    def make_gt(self, image_shape, ann, ignore_not_mapped_classes=False):
        # int32 instead of int64 because opencv cannot draw on int64 bitmaps.
        gt = np.ones(image_shape[:2], dtype=np.int32) * self._bkg_color  # default bkg
        for label in ann.labels:
            gt_color = self._class_mapping.get(label.obj_class.name, None)
            if gt_color is None:
                # TODO: Factor out this check to lib
                if not ignore_not_mapped_classes:
                    raise RuntimeError('Missing class mapping (title to index). Class {}.'.format(label.obj_class.name))
            else:
                label.geometry.draw(gt, gt_color)
        gt = sly_image.resize_inter_nearest(gt, self._out_size).astype(np.int64)
        return gt

    def _get_sample_impl(self, img_fpath, ann_fpath):
        img = sly_image.read(img_fpath)
        ann = self.load_annotation(ann_fpath)
        gt = self.make_gt(img.shape, ann)
        img = sly_image.resize(img, self._out_size)
        return img, gt

    def __getitem__(self, index):
        for att in range(self._catcher_retries):
            img_path, ann_path = self._samples[index]
            res = self._sample_catcher.exec(index,
                                            {'img': img_path,
                                             'ann': ann_path},
                                            self._get_sample_impl,
                                            img_path,
                                            ann_path)
            if res is not None:
                return res
            index = random.randrange(len(self._samples))  # must be ok for large ds
        raise RuntimeError('Unable to load some correct sample.')


def partition_train_val(num_samples, val_fraction):
    """Returns a bool array indicating whether a given sample falls into a train fold (otherwise samples fall to val).
    """

    if num_samples <= 1:
        raise ValueError('Need at least 2 samples to prepare a training/validation split (at least 1 each for training '
                         'and validation).')

    val_boundary = round(num_samples * val_fraction)
    # Make sure there is at least one sample in both training and validation fold,
    # adjust boundary if necessary.
    if val_boundary <= 0:
        val_boundary = 1
    elif val_boundary >= num_samples:
        val_boundary = num_samples - 1
    is_train_sample = ([False] * val_boundary) + ([True] * (num_samples - val_boundary))
    random.shuffle(is_train_sample)
    return is_train_sample



from typing import List, Dict, Optional

from supervisely.app.widgets import Widget
from supervisely.app.content import StateJson
from supervisely.api.user_api import UserInfo


class MembersListPreview(Widget):
    """Displays a scrollable list of team members (login, role) from UserInfo objects."""

    def __init__(
        self,
        users: Optional[List[UserInfo]] = [],
        max_width: int = 300,
        empty_text: str = None,
        widget_id: int = None,
    ):
        """
        :param users: List of UserInfo to display.
        :type users: List[UserInfo], optional
        :param max_width: Max width in pixels (min 150).
        :type max_width: int
        :param empty_text: Text when list is empty.
        :type empty_text: str, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: int, optional
        """
        self._users = users
        self._max_width = self._get_max_width(max_width)
        self._empty_text = empty_text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_max_width(self, value):
        if value < 150:
            value = 150
        return f"{value}px"

    def get_json_data(self):
        return {
            "maxWidth": self._max_width,
        }

    def get_json_state(self) -> Dict:
        return {
            "users": [
                {
                    "login": f"<b>{user.login}</b>",
                    "role": user.role,
                }
                for user in self._users
            ]
        }

    def set(self, users: List[UserInfo]):
        self._users = users
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from supervisely import Api
from supervisely.api.app_api import AppInfo
from supervisely.app import DataJson
from supervisely.app.widgets import Button, Card, Stepper, Text, Widget

button_clicked = {}


def update_custom_params(
    button: Button,
    params_dct: Dict[str, Any],
) -> None:
    button_state = button.get_json_data()
    for key in params_dct.keys():
        if key not in button_state:
            raise AttributeError(f"Parameter {key} doesn't exists.")
        else:
            DataJson()[button.widget_id][key] = params_dct[key]
    DataJson().send_changes()


def update_custom_button_params(
    button: Button,
    params_dct: Dict[str, Any],
) -> None:
    params = params_dct.copy()
    if "icon" in params and params["icon"] is not None:
        new_icon = f'<i class="{params["icon"]}" style="margin-right: {button._icon_gap}px"></i>'
        params["icon"] = new_icon
    update_custom_params(button, params)


def disable_enable(widgets: List[Widget], disable: bool = True):
    for w in widgets:
        if disable:
            w.disable()
        else:
            w.enable()


def unlock_lock(cards: List[Card], unlock: bool = True, message: str = None):
    for w in cards:
        if unlock:
            w.unlock()
            # w.uncollapse()
        else:
            w.lock(message)
            # w.collapse()


def collapse_uncollapse(cards: List[Card], collapse: bool = True):
    for w in cards:
        if collapse:
            w.collapse()
        else:
            w.uncollapse()


def wrap_button_click(
    button: Button,
    cards_to_unlock: List[Card],
    widgets_to_disable: List[Widget],
    callback: Optional[Callable] = None,
    lock_msg: str = None,
    upd_params: bool = True,
    validation_text: Text = None,
    validation_func: Optional[Callable] = None,
    on_select_click: Optional[Callable] = None,
    on_reselect_click: Optional[Callable] = None,
    collapse_card: Tuple[Card, bool] = None,
) -> Callable[[Optional[bool]], None]:
    global button_clicked

    select_params = {"icon": None, "plain": False, "text": "Select"}
    reselect_params = {"icon": "zmdi zmdi-refresh", "plain": True, "text": "Reselect"}
    bid = button.widget_id
    button_clicked[bid] = False

    def button_click(button_clicked_value: Optional[bool] = None):
        if button_clicked_value is None or button_clicked_value is False:
            if validation_func is not None:
                success = validation_func()
                if not success:
                    return

        if button_clicked_value is not None:
            button_clicked[bid] = button_clicked_value
        else:
            button_clicked[bid] = not button_clicked[bid]

        if button_clicked[bid] and upd_params:
            update_custom_button_params(button, reselect_params)
            if on_select_click is not None:
                for func in on_select_click:
                    func()
        else:
            update_custom_button_params(button, select_params)
            if on_reselect_click is not None:
                for func in on_reselect_click:
                    func()
            validation_text.hide()

        unlock_lock(
            cards_to_unlock,
            unlock=button_clicked[bid],
            message=lock_msg,
        )
        disable_enable(
            widgets_to_disable,
            disable=button_clicked[bid],
        )
        if callback is not None and not button_clicked[bid]:
            callback(False)

        if collapse_card is not None:
            card, collapse = collapse_card
            if collapse:
                collapse_uncollapse([card], collapse)

    return button_click


def set_stepper_step(stepper: Stepper, button: Button, next_pos: int):
    bid = button.widget_id
    if button_clicked[bid] is True:
        stepper.set_active_step(next_pos)
    else:
        stepper.set_active_step(next_pos - 1)


def get_module_info_by_name(api: Api, app_name: str) -> Union[Dict, None]:
    all_modules = api.app.get_list_ecosystem_modules()
    for module in all_modules:
        if module["name"] == app_name:
            app_info = api.app.get_info(module["id"])
            return app_info


def generate_task_check_function_js(folder: str) -> str:
    """
    Returns JavaScript function code for checking existing tasks.

    :param folder: Remote folder to check.
    :type folder: str
    :returns: JavaScript function code for checking existing tasks.
    :rtype: str
    """
    escaped_folder = folder.replace("'", "\\'")
    js_code = f"""
        if (!task || !task.meta || !task.meta.params || !task.meta.params.state) {{
            return false;
        }}
        const taskFolder = task.meta.params.state.slyFolder;
        if (!taskFolder || typeof taskFolder !== 'string') {{
            return false;
        }}
        return taskFolder === '{escaped_folder}';
    """
    return js_code



from typing import Any, Callable, Dict, List, Literal, Optional, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Empty, Widget


class ColorPicker(Widget):
    """Color selector supporting multiple formats (hex, rgb, etc.).
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/input/colorpicker>`_.
    """

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        show_alpha: Optional[bool] = False,
        color_format: Optional[Literal["hex", "hsl", "hsv", "rgb"]] = "hex",
        compact: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """
        :param show_alpha: if True, alpha channel will be shown
        :type show_alpha: bool, optional
        :param color_format: color format, one of: hex, hsl, hsv, rgb
        :type color_format: Literal["hex", "hsl", "hsv", "rgb"]
        :param compact: if True, compact mode will be enabled
        :type compact: bool, optional
        :param widget_id: An identifier of the widget.
        :type widget_id: str, optional
        :raises TypeError: if color_format is not one of hex, hsl, hsv, rgb

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import ColorPicker

                color_picker = ColorPicker(color_format="rgb")
        """
        self._show_alpha = show_alpha
        self._color_format = color_format
        self._changes_handled = False
        self._color_info = Empty()
        self._compact = compact

        if self._color_format not in ["hsl", "hsv", "hex", "rgb"]:
            raise TypeError(
                f"Incorrect color format: {self._color_format}, only hsl, hsv, hex, rgb are possible"
            )

        if self._color_format == "hex":
            self._color = "#20a0ff"
        elif self._color_format == "hsl":
            self._color = "hsl(205, 100%, 56%)"
        elif self._color_format == "hsv":
            self._color = "hsv(205, 87%, 100%)"
        else:
            self._color = "rgb(32, 160, 255)"

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[bool, str]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - show_alpha: if True, alpha channel will be shown
            - color_format: color format, one of: hex, hsl, hsv, rgb
            - compact: if True, compact mode will be enabled

        :returns: dictionary with widget data
        :rtype: Dict[str, Union[bool, str]]
        """
        return {
            "show_alpha": self._show_alpha,
            "color_format": self._color_format,
            "compact": self._compact,
        }

    def get_json_state(self) -> Dict[str, Union[str, List[int]]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - color: current color

        :returns: dictionary with widget state
        :rtype: Dict[str, Union[str, List[int]]]
        """
        return {"color": self._color}

    def get_value(self) -> Union[str, List[int]]:
        """Returns current color.

        :returns: current color
        :rtype: Union[str, List[int]]
        """
        return StateJson()[self.widget_id]["color"]

    def set_value(self, value: Optional[Union[str, List[int]]]) -> None:
        """Sets current color.

        :param value: current color
        :type value: Union[str, List[int]]]
        """
        self._color = value
        if isinstance(self._color, list) and len(self._color) == 3 and self._color_format == "rgb":
            if (
                isinstance(self._color[0], int)
                and isinstance(self._color[1], int)
                and isinstance(self._color[2], int)
            ):
                self._color = f"rgb({self._color[0]}, {self._color[1]}, {self._color[2]})"
        if (
            (self._color_format == "hex" and self._color[0] != "#")
            or (self._color_format == "hsl" and self._color[0:3] != "hsl")
            or (self._color_format == "hsv" and self._color[0:3] != "hsv")
            or (self._color_format == "rgb" and self._color[0:3] != "rgb")
        ):
            raise ValueError(
                f"Incorrect input value format: {self._color}, {self._color_format} format should be, check your input data"
            )
        StateJson()[self.widget_id]["color"] = self._color
        StateJson().send_changes()

    def is_show_alpha_enabled(self) -> bool:
        """Returns True if alpha channel is shown, False otherwise.

        :returns: True if alpha channel is shown, False otherwise
        :rtype: bool
        """
        return DataJson()[self.widget_id]["show_alpha"]

    def disable_show_alpha(self) -> None:
        """Disables alpha channel."""
        self._show_alpha = False
        DataJson()[self.widget_id]["show_alpha"] = self._show_alpha
        DataJson().send_changes()

    def enable_show_alpha(self) -> None:
        """Enables alpha channel."""
        self._show_alpha = True
        DataJson()[self.widget_id]["show_alpha"] = self._show_alpha
        DataJson().send_changes()

    def get_color_format(self) -> Literal["hex", "hsl", "hsv", "rgb"]:
        """Returns current color format.

        :returns: current color format
        :rtype: Literal["hex", "hsl", "hsv", "rgb"]
        """
        return DataJson()[self.widget_id]["color_format"]

    def set_color_format(self, value: Literal["hex", "hsl", "hsv", "rgb"]) -> None:
        """Sets current color format.

        :param value: current color format
        :type value: Literal["hex", "hsl", "hsv", "rgb"]
        """
        self._color_format = value
        DataJson()[self.widget_id]["color_format"] = self._color_format
        DataJson().send_changes()

    def value_changed(self, func: Callable[[Union[str, List[int]]], Any]) -> Callable[[], None]:
        """Decortator for function that will be called when color is changed.

        :param func: function that will be called when color is changed
        :type func: Callable[[Union[str, List[int]]], Any]
        :returns: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(ColorPicker.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._color = res
            func(res)

        return _click



# coding: utf-8
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.volume_annotation.volume_object import VolumeObject


class VolumeObjectCollection(VideoObjectCollection):
    """
    Collection with :class:`~supervisely.volume_annotation.volume_object.VolumeObject` instances. :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection` object is immutable.

    :Usage Example:

        .. code-block:: python

            import supervisely as sly

            # Create two VolumeObjects for collection
            class_heart = sly.ObjClass('heart', sly.Rectangle)
            obj_heart = sly.VolumeObject(class_heart)
            class_lang = sly.ObjClass('lang', sly.Rectangle)
            obj_lang = sly.VolumeObject(class_lang)

            # Create VolumeObjectCollection
            obj_collection = sly.VolumeObjectCollection([obj_heart, obj_lang])
            obj_collection_json = obj_collection.to_json()
            print(obj_collection_json)
            # Output: [
            #     {
            #         "key": "773300c59fde4707887068d555269ba5",
            #         "classTitle": "heart",
            #         "tags": []
            #     },
            #     {
            #         "key": "8257371497d2402cadabc690f796b1d1",
            #         "classTitle": "lang",
            #         "tags": []
            #     }
            # ]

            # Add item to VolumeObjectCollection
            class_arm = sly.ObjClass('arm', sly.Rectangle)
            obj_arm = sly.VolumeObject(class_arm)
            # Remember that VolumeObjectCollection is immutable, and we need to assign new instance of VolumeObjectCollection to a new variable
            new_obj_collection = obj_collection.add(obj_arm)
            new_obj_collection_json = new_obj_collection.to_json()
            print(new_obj_collection_json)
            # Output: [
            #     {
            #         "key": "1b62882f180f49ae96744e13fba26d02",
            #         "classTitle": "heart",
            #         "tags": []
            #     },
            #     {
            #         "key": "0c95107fce6c4ac1b42132e83c1a6b3b",
            #         "classTitle": "lang",
            #         "tags": []
            #     },
            #     {
            #         "key": "660d8572e20c4101b4cc7edf7f04b090",
            #         "classTitle": "arm",
            #         "tags": []
            #     }
            # ]

            # You can also add multiple items to collection
            class_arm = sly.ObjClass('arm', sly.Rectangle)
            obj_arm = sly.VolumeObject(class_arm)
            class_train = sly.ObjClass('train', sly.Rectangle)
            obj_train = sly.VolumeObject(class_train)
            # Remember that VolumeObjectCollection is immutable, and we need to assign new instance of VolumeObjectCollection to a new variable
            new_obj_collection = obj_collection.add_items([obj_arm, obj_train])
            new_obj_collection_json = new_obj_collection.to_json()
            print(new_obj_collection_json)
            # Output: [
            #     {
            #         "key": "892c62aae48b495687fe8b33ce8ebe96",
            #         "classTitle": "heart",
            #         "tags": []
            #     },
            #     {
            #         "key": "db0f64a7b60447769374943077e57679",
            #         "classTitle": "lang",
            #         "tags": []
            #     },
            #     {
            #         "key": "72c24a107f344ef68d4fe8e93dff8184",
            #         "classTitle": "arm",
            #         "tags": []
            #     },
            #     {
            #         "key": "d20738a608234152b43b1c23b7958c47",
            #         "classTitle": "train",
            #         "tags": []
            #     }
            # ]

            # Find intersection of given list of instances with collection items
            obj_intersections = obj_collection.intersection([obj_heart])
            obj_intersections_json = obj_intersections.to_json()
            print(obj_intersections_json)
            # Output: [
            #     {
            #         "key": "5fae14a4a42c42a29904a887442162c9",
            #         "classTitle": "heart",
            #         "tags": []
            #     }
            # ]

            # Find difference between collection and given list of VolumeObject
            obj_diff = obj_collection.difference([obj_heart])
            obj_diff_json = obj_diff.to_json()
            print(obj_diff_json)
            # Output: [
            #     {
            #         "key": "b4365bbec0314de58e20267735a39164",
            #         "classTitle": "lang",
            #         "tags": []
            #     }
            # ]

            # Merge collection and given list of VolumeObjectCollection

            class_arm = sly.ObjClass('arm', sly.Rectangle)
            obj_arm = sly.VolumeObject(class_arm)
            class_train = sly.ObjClass('train', sly.Rectangle)
            obj_train = sly.VolumeObject(class_train)
            over_collection = sly.VolumeObjectCollection([obj_arm, obj_train])
            # Merge
            merge_collection = obj_collection.merge(over_collection)
            merge_collection_json = merge_collection.to_json()
            print(merge_collection_json)
            # Output: [
            #     {
            #         "key": "8e6f9f05f0954193bdb2005b9cd6e20b",
            #         "classTitle": "arm",
            #         "tags": []
            #     },
            #     {
            #         "key": "a010f91a5e7646bdb8e06dfd6fa50f30",
            #         "classTitle": "train",
            #         "tags": []
            #     },
            #     {
            #         "key": "8440e915a94640b0be279c4fa293553b",
            #         "classTitle": "heart",
            #         "tags": []
            #     },
            #     {
            #         "key": "e03e7dd2a2d14854945b6ee7bf7c72e3",
            #         "classTitle": "lang",
            #         "tags": []
            #     }
            # ]
    """

    item_type = VolumeObject



import os
import nrrd
import math
import supervisely
import trimesh
import numpy as np

from typing import List, Dict
from supervisely import logger
from supervisely.io.fs import (
    get_file_name_with_ext,
    file_exists,
    dir_exists,
    mkdir,
    change_directory_at_index,
)

if not hasattr(np, "bool"):
    np.bool = np.bool_

def matrix_from_nrrd_header(header: Dict) -> np.ndarray:
    """
    Creates matrix from NRRD header

    :param header: Dictionary containing the header fields and their corresponding parsed value
    :type header: dict (str, :obj:`Object`)
    :returns: Transformation matrix that maps voxel coordinates to world coordinates
    :rtype: np.ndarray
    :raises IOError: if Need "{}" field of header to determine the mapping from voxels to world coordinates
    """
    try:
        space_directions = header["space directions"]
        space_origin = header["space origin"]
    except KeyError as e:
        raise IOError(
            'Need the header\'s "{}" field to determine the mapping from voxels to world coordinates.'.format(
                e
            )
        )

    # "... the space directions field gives, one column at a time, the mapping from image space to world space
    # coordinates ... [1]_" -> list of columns, needs to be transposed
    trans_3x3 = np.array(space_directions).T
    trans_4x4 = np.eye(4)
    trans_4x4[:3, :3] = trans_3x3
    trans_4x4[:3, 3] = space_origin

    return trans_4x4


def voxels_to_mask(mask_shape: List, voxel_to_world: np.ndarray, stl_path: str) -> np.ndarray:
    """
    Converts STL voxels to NRRD mask

    :param mask_shape: Value "sizes" of Volume nrrd_header
    :type mask_shape: List
    :param voxel_to_world: Transformation matrix that maps voxel coordinates to world coordinates
    :type voxel_to_world: np.ndarray
    :param stl_path: Path to STL file
    :type stl_path: str
    :returns: Mask
    :rtype: np.ndarray
    """

    world_to_voxel = np.linalg.inv(voxel_to_world)

    mesh = trimesh.load(stl_path)

    min_vec = [float("inf"), float("inf"), float("inf")]
    max_vec = [float("-inf"), float("-inf"), float("-inf")]

    mesh.apply_scale((-1, -1, 1))  # LPS to RAS
    mesh.apply_transform(world_to_voxel)

    for vert in mesh.vertices:
        min_vec[0] = min(min_vec[0], vert[0])
        min_vec[1] = min(min_vec[1], vert[1])
        min_vec[2] = min(min_vec[2], vert[2])

        max_vec[0] = max(max_vec[0], vert[0])
        max_vec[1] = max(max_vec[1], vert[1])
        max_vec[2] = max(max_vec[2], vert[2])

    center = [(min_v + max_v) / 2 for min_v, max_v in zip(min_vec, max_vec)]

    try:
        voxel = mesh.voxelized(pitch=1.0)
    except Exception as e:
        supervisely.logger.error(e)
        supervisely.logger.warning(
            "Couldn't voxelize file {!r}".format(get_file_name_with_ext(stl_path)),
            extra={"file_path": stl_path},
        )
        return np.zeros(mask_shape).astype(np.bool)

    voxel = voxel.fill()
    mask = voxel.matrix.astype(np.bool)
    padded_mask = np.zeros(mask_shape).astype(np.bool)

    # find dimension coords
    start = [math.ceil(center_v - shape_v / 2) for center_v, shape_v in zip(center, mask.shape)]
    end = [math.ceil(center_v + shape_v / 2) for center_v, shape_v in zip(center, mask.shape)]

    # find intersections
    vol_inter_max = [max(start[0], 0), max(start[1], 0), max(start[2], 0)]
    vol_inter_min = [
        min(end[0], mask_shape[0]),
        min(end[1], mask_shape[1]),
        min(end[2], mask_shape[2]),
    ]

    padded_mask[
        vol_inter_max[0] : vol_inter_min[0],
        vol_inter_max[1] : vol_inter_min[1],
        vol_inter_max[2] : vol_inter_min[2],
    ] = mask[
        vol_inter_max[0] - start[0] : vol_inter_min[0] - start[0],
        vol_inter_max[1] - start[1] : vol_inter_min[1] - start[1],
        vol_inter_max[2] - start[2] : vol_inter_min[2] - start[2],
    ]
    padded_mask = padded_mask.astype(dtype="uint8")
    return padded_mask


def to_nrrd(
    stl_paths: List[str], nrrd_paths: List[str], volume_path: str = None, header: Dict = None
) -> None:
    """
    Save STL files as NRRD files containing 3D Masks in the same directory with the same name.

    :param stl_paths: Paths to STL files with interpolation.
    :type stl_paths: List[str]
    :param nrrd_paths: Paths to NRRD files with 3D Masks.
    :type nrrd_paths: List[str]
    :param volume_path: Path to the Volume NRRD file, from which we obtain the header. Has priority if defined together with the header.
    :type volume_path: str
    :param header: Dictionary with NRRD volume header parameters, must be used when there is no Volume NRRD file available.
    :type header: Dict
    :returns: None
    :rtype: NoneType
    """

    if len(stl_paths) != len(nrrd_paths):
        raise ValueError("The lengths of the lists 'stl_paths' and 'nrrd_paths' do not match")

    make_warn = False

    for stl_path, nrrd_path in zip(stl_paths, nrrd_paths):
        # no need to convert if a converted interpolation in NRRD format already exists
        if file_exists(nrrd_path):
            continue
        else:
            nrrd_dir = os.path.dirname(nrrd_path)
            if not dir_exists(nrrd_dir):
                mkdir(nrrd_dir)

        if volume_path is None:
            # trying to find the Volume file inside the project
            volume_path = os.path.dirname(change_directory_at_index(stl_path, "volume", -3))

        if file_exists(volume_path) is False and header is None:
            raise FileNotFoundError(
                "NRRD Volume file not found and header not set, unable to convert STL"
            )
        elif file_exists(volume_path) is True:
            try:
                _, header = nrrd.read(volume_path)
            except Exception as e:
                e.args = (
                    "Unable to retrieve header from the NRRD Volume file. File is empty or corrupted",
                )
        world_matrix = matrix_from_nrrd_header(header)
        shape = header["sizes"]
        mask = voxels_to_mask(shape, world_matrix, stl_path)
        nrrd.write(nrrd_path, mask, header)
        make_warn = True
    if make_warn:
        logger.warning(
            f"STL format is no longer supported. STL is automatically converted to NRRD containing 3D Mask"
        )



import queue
import asyncio
from typing import Any, Optional, List
from enum import Enum


class RequestType(Enum):
    """Types of requests handled by the live training request queue."""

    START = "start"
    PREDICT = "predict"
    ADD_SAMPLE = "add-sample"
    STATUS = "status"


class Request:
    """A simple representation of an API request."""

    def __init__(self, request_type: RequestType, data: Optional[dict] = None, future: Optional[asyncio.Future] = None):
        """
        :param request_type: RequestType enum.
        :type request_type: RequestType
        :param data: Optional request data.
        :type data: dict
        :param future: Optional Future for response.
        :type future: asyncio.Future
        """
        self.type = request_type
        self.data = data
        self.future = future

    def to_tuple(self):
        return (self.type, self.data, self.future)


class RequestQueue:
    """Thread-safe queue for API requests."""
    
    def __init__(self):
        self._queue = queue.Queue()
    
    def put(self, request_type: RequestType, data: Optional[dict] = None) -> asyncio.Future:
        """Add request and return future for result."""
        future = asyncio.Future()
        self._queue.put(Request(request_type, data, future))
        return future
    
    def get_all(self) -> List[Request]:
        """Get all pending requests (non-blocking)."""
        requests = []
        while not self._queue.empty():
            try:
                requests.append(self._queue.get_nowait())
            except queue.Empty:
                break
        return requests
    
    def is_empty(self) -> bool:
        return self._queue.empty()
    
    def get(self, timeout: float = None) -> Request:
        """Get a single request from the queue."""
        return self._queue.get(timeout=timeout)



# coding: utf-8
import os
import shutil
import tarfile

import requests

from supervisely.api.api import Api
from supervisely.io.fs import (
    ensure_base_path,
    get_file_name,
    get_subdirs,
    remove_dir,
    silent_remove,
)
from supervisely.task.progress import Progress


def download(github_url, dest_dir, github_token=None, version="master", log_progress=True):
    tar_path = os.path.join(dest_dir, 'repo.tar.gz')
    download_tar(github_url, tar_path, github_token, version, log_progress)

    with tarfile.open(tar_path) as archive:
        archive.extractall(dest_dir)

    subdirs = get_subdirs(dest_dir)
    if len(subdirs) != 1:
        raise RuntimeError("Repo is downloaded and extracted, but resulting directory not found")
    extracted_path = os.path.join(dest_dir, subdirs[0])

    for filename in os.listdir(extracted_path):
        shutil.move(os.path.join(extracted_path, filename), os.path.join(dest_dir, filename))
    remove_dir(extracted_path)
    silent_remove(tar_path)


def download_tar(
    github_url: str,
    tar_path: str,
    github_token: str = None,
    version: str = "master",
    log_progress: bool = True,
) -> None:
    """
    Download a tarball from a GitHub repository.

    :param github_url: Github repository URL
    :type github_url: str
    :param tar_path: Path to save the downloaded tar file
    :type tar_path: str
    :param github_token: Github token for authentication
    :type github_token: str
    :param version: Branch or tag to download
    :type version: str
    :param log_progress: Whether to log progress during download
    :type log_progress: bool
    """
    headers = {}
    if github_token is not None:
        headers = {"Authorization": "token {}".format(github_token)}

    ensure_base_path(tar_path)

    if ".git" not in github_url:
        github_url += ".git"
    tar_url = github_url.replace(".git", "/archive/{}.tar.gz".format(version))
    r = requests.get(tar_url, headers=headers, stream=True)
    if r.status_code != requests.codes.ok:  # pylint: disable=no-member
        Api._raise_for_status(r)

    if log_progress:
        progress = Progress("Downloading (KB)", len(r.content) / 1024)
    with open(tar_path, 'wb') as f:
        for chunk in r.iter_content(chunk_size=8192):
            f.write(chunk)
            if log_progress:
                progress.iters_done_report(len(chunk) / 1024)



from supervisely import Annotation, ProjectMeta
from supervisely.convert.base_converter import AvailableVideoConverters, BaseConverter
from supervisely.convert.video.video_converter import VideoConverter


class MOTConverter(VideoConverter):
    """Base for MOT (Multiple Object Tracking) challenge format import; to_supervisely not implemented."""

    def __str__(self):
        return AvailableVideoConverters.MOT

    def validate_format(self):
        return False

    def to_supervisely(self, item: VideoConverter.Item, meta: ProjectMeta) -> Annotation:
        raise NotImplementedError()



# coding: utf-8
"""Base classes and helpers for Supervisely API modules."""

import asyncio
from collections import namedtuple
from copy import deepcopy
from math import ceil
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
)

import requests

from supervisely._utils import batched, camel_to_snake

if TYPE_CHECKING:
    from supervisely.api.api import Api


class ApiField:
    """Fields for API requests."""

    ID = "id"
    """"""
    NAME = "name"
    """"""
    DESCRIPTION = "description"
    """"""
    CREATED_AT = "createdAt"
    """"""
    UPDATED_AT = "updatedAt"
    """"""
    ROLE = "role"
    """"""
    TEAM_ID = "teamId"
    """"""
    WORKSPACE_ID = "workspaceId"
    """"""
    CONFIG = "config"
    """"""
    SIZE = "size"
    """"""
    PLUGIN_ID = "pluginId"
    """"""
    PLUGIN_VERSION = "pluginVersion"
    """"""
    HASH = "hash"
    """"""
    STATUS = "status"
    """"""
    ONLY_TRAIN = "onlyTrain"
    """"""
    USER_ID = "userId"
    """"""
    README = "readme"
    """"""
    PROJECT_ID = "projectId"
    """"""
    IMAGES_COUNT = "imagesCount"
    """"""
    ITEMS_COUNT = "itemsCount"
    """"""
    DATASET_ID = "datasetId"
    """"""
    LINK = "link"
    """"""
    WIDTH = "width"
    """"""
    HEIGHT = "height"
    """"""
    WEIGHTS_LOCATION = "weightsLocation"
    """"""
    IMAGE_ID = "imageId"
    """"""
    IMAGE_NAME = "imageName"
    """"""
    ANNOTATION = "annotation"
    """"""
    TASK_ID = "taskId"
    """"""
    LABELS_COUNT = "labelsCount"
    """"""
    FILTER = "filter"
    """"""
    FILTERS = "filters"
    """"""
    META = "meta"
    """"""
    FILE_META = "fileMeta"
    """"""
    SHARED_LINK = "sharedLinkToken"
    """"""
    EXPLORE_PATH = "explorePath"
    """"""
    MIME = "mime"
    """"""
    EXT = "ext"
    """"""
    TYPE = "type"
    """"""
    DEFAULT_VERSION = "defaultVersion"
    """"""
    DOCKER_IMAGE = "dockerImage"
    """"""
    CONFIGS = "configs"
    """"""
    VERSIONS = "versions"
    """"""
    VERSION = "version"
    """"""
    TOKEN = "token"
    """"""
    CAPABILITIES = "capabilities"
    """"""
    STARTED_AT = "startedAt"
    """"""
    FINISHED_AT = "finishedAt"
    """"""
    AGENT_ID = "agentId"
    """"""
    MODEL_ID = "modelId"
    """"""
    RESTART_POLICY = "restartPolicy"
    """"""
    SETTINGS = "settings"
    """"""
    SORT = "sort"
    """"""
    SORT_ORDER = "sort_order"
    """"""
    IMAGES = "images"
    """"""
    IMAGE_IDS = "imageIds"
    """"""
    ANNOTATIONS = "annotations"
    """"""
    EMAIL = "email"
    """"""
    LOGIN = "login"
    """"""
    LOGINS = "logins"
    """"""
    DISABLED = "disabled"
    """"""
    DISABLED_AT = "disabledAt"
    """"""
    DISABLED_BY = "disabledBy"
    """"""
    LAST_LOGIN = "lastLogin"
    """"""
    PASSWORD = "password"
    """"""
    ROLE_ID = "roleId"
    """"""
    IS_RESTRICTED = "isRestricted"
    """"""
    DISABLE = "disable"
    """"""
    TEAMS = "teams"
    """"""
    USER_IDS = "userIds"
    """"""
    PROJECT_NAME = (["projectTitle"], "project_name")
    """"""
    DATASET_NAME = (["datasetTitle"], "dataset_name")
    """"""
    WORKSPACE_NAME = (["workspaceTitle"], "workspace_name")
    """"""
    CREATED_BY_ID = (["createdBy"], "created_by_id")
    """"""
    CREATED_BY_LOGIN = (["managerLogin"], "created_by_login")
    """"""
    CREATED_BY_USER = "createdByUser"
    """"""
    ASSIGNED_TO_ID = (["userId"], "assigned_to_id")
    """"""
    ASSIGNED_TO_LOGIN = (["labelerLogin"], "assigned_to_login")
    """"""
    FINISHED_IMAGES_COUNT = "finishedImagesCount"
    """"""
    PROGRESS_IMAGES_COUNT = "progressImagesCount"
    """"""
    CLASSES_TO_LABEL = (["meta", "classes"], "classes_to_label")
    """"""
    TAGS_TO_LABEL = (["meta", "projectTags"], "tags_to_label")
    """"""
    IMAGES_RANGE = (["meta", "range"], "images_range")
    """"""
    REJECTED_IMAGES_COUNT = "rejectedImagesCount"
    """"""
    ACCEPTED_IMAGES_COUNT = "acceptedImagesCount"
    """"""
    OBJECTS_LIMIT_PER_IMAGE = (["meta", "imageFiguresLimit"], "objects_limit_per_image")
    """"""
    TAGS_LIMIT_PER_IMAGE = (["meta", "imageTagsLimit"], "tags_limit_per_image")
    """"""
    FILTER_IMAGES_BY_TAGS = (["meta", "imageTags"], "filter_images_by_tags")
    """"""
    INCLUDE_IMAGES_WITH_TAGS = ([], "include_images_with_tags")
    """"""
    EXCLUDE_IMAGES_WITH_TAGS = ([], "exclude_images_with_tags")
    """"""
    SIZEB = (["size"], "sizeb")
    """"""
    FRAMES = "frames"
    """"""
    FRAMES_COUNT = (["fileMeta", "framesCount"], "frames_count")
    """"""
    PATH_ORIGINAL = "pathOriginal"
    """"""
    OBJECTS_COUNT = "objectsCount"
    """"""
    FRAMES_TO_TIMECODES = (["fileMeta", "framesToTimecodes"], "frames_to_timecodes")
    """"""
    TAGS = "tags"
    """"""
    VIDEO_ID = "videoId"
    """"""
    FRAME_INDEX = (["meta", "frame"], "frame_index")
    """"""
    LABELING_TOOL = "tool"
    """"""
    GEOMETRY_TYPE = "geometryType"
    """"""
    GEOMETRY = "geometry"
    """"""
    GEOMETRY_META = "geometryMeta"
    """"""
    OBJECT_ID = "objectId"
    """"""
    ANNOTATION_OBJECT_ID = "annotationObjectId"
    """"""
    FRAME = "frame"
    """"""
    # FIGURES_COUNT = (['labelsCount'], 'figures_count')
    """"""
    STREAMS = "streams"
    """"""
    VIDEO_IDS = "videoIds"
    """"""
    FRAME_WIDTH = (["fileMeta", "width"], "frame_width")
    """"""
    FRAME_HEIGHT = (["fileMeta", "height"], "frame_height")
    """"""
    VIDEO_NAME = "videoName"
    """"""
    FRAME_RANGE = "frameRange"
    """"""
    TRACK_ID = "trackId"
    """"""
    PROGRESS = "progress"
    """"""
    CURRENT = "current"
    """"""
    TOTAL = "total"
    """"""
    STOPPED = "stopped"
    """"""
    VIDEOS = "videos"
    """"""
    FILENAME = "filename"
    """"""
    SHAPE = "shape"
    """"""
    COLOR = "color"
    """"""
    CLASS_ID = "classId"
    """"""
    ENTITY_ID = "entityId"
    """"""
    ANNOTATION_OBJECTS = "annotationObjects"
    """"""
    TAG_ID = "tagId"
    """"""
    TAG_IDS = "tagIds"
    """"""
    ERROR = "error"
    """"""
    MESSAGE = "message"
    """"""
    CONTENT = "content"
    """"""
    FIGURES = "figures"
    """"""
    LAYOUT = "layout"
    """"""
    WIDGETS = "widgets"
    """"""
    CLOUD_MIME = (["fileMeta", "mime"], "cloud_mime")
    """"""
    PREVIEW = "preview"
    """"""
    FIGURES_COUNT = "figuresCount"
    """"""
    ANN_OBJECTS_COUNT = (["annotationObjectsCount"], "objects_count")
    """"""
    POINTCLOUD_ID = "pointCloudId"
    """"""
    POINTCLOUD_IDS = "pointCloudIds"
    """"""
    POINTCLOUDS = "pointClouds"
    """"""
    ADVANCED = "advanced"
    """"""
    IGNORE_AGENT = "ignoreAgent"
    """"""
    SCRIPT = "script"
    """"""
    LOGS = "logs"
    """"""
    FILES = "files"
    """"""
    HASHES = "hashes"
    """"""
    SUBTITLE = "subtitle"
    """"""
    COMMAND = "command"
    """"""
    DEFAULT_VALUE = "defaultValue"
    """"""
    TITLE = "title"
    """"""
    AREA = "area"
    """"""
    OPTIONS = "options"
    """"""
    REPORT_ID = "reportId"
    """"""
    WIDGET = "widget"
    """"""
    PAYLOAD = "payload"
    """"""
    FIELD = "field"
    """"""
    FIELDS = "fields"
    """"""
    APPEND = "append"
    """"""
    WITH_CUSTOM_DATA = "withCustomBigData"
    """"""
    PATH = "path"
    """"""
    SESSION_ID = "sessionId"
    """"""
    ACTION = "action"
    """"""
    FIGURE_ID = "figureId"
    """"""
    FIGURE_IDS = "figureIds"
    """"""
    VALUE = "value"
    """"""
    ZOOM_FACTOR = "zoomFactor"
    """"""
    FULL_STORAGE_URL = "fullStorageUrl"
    """"""
    REVIEWER_ID = "reviewerId"
    """"""
    REVIEWER_LOGIN = "reviewerLogin"
    """"""
    RECURSIVE = "recursive"
    """"""
    ECOSYSTEM_ITEM_ID = "moduleId"
    """"""
    APP_ID = "appId"
    """"""
    PROJECT = "project"
    """"""
    OUTPUT = "output"
    """"""
    REFERENCE_IMAGE_URL = "referenceImageUrl"
    """"""
    GENERAL = "general"
    """"""
    ENTITIES = "entities"
    """"""
    STORAGE_PATH = "storagePath"
    """"""
    EXT2 = (["meta", "ext"], "ext")
    """"""
    MIME2 = (["meta", "mime"], "mime")
    """"""
    SIZEB2 = (["meta", "size"], "sizeb")
    """"""
    JOB_ID = "jobId"
    """"""
    DATASETS_COUNT = "datasetsCount"
    """"""
    CUSTOM_DATA = "customData"
    """"""
    CONTEXT = "context"
    """"""
    STATE = "state"
    """"""
    IDS = "ids"
    """"""
    DATE = "date"
    """"""
    PARAMS = "params"
    """"""
    LOG_LEVEL = "logLevel"
    """"""
    APP_VERSION = "appVersion"
    """"""
    IS_BRANCH = "isBranch"
    """"""
    TASK_NAME = "taskName"
    """"""
    PROXY_KEEP_URL = "proxyKeepUrl"
    """"""
    USERS_IDS = "usersIds"
    """"""
    MODULE_ID = "moduleId"
    """"""
    USER_LOGIN = "userLogin"
    """"""
    SLUG = "slug"
    """"""
    IS_SHARED = "isShared"
    """"""
    TASKS = "tasks"
    """"""
    REPO = "repo"
    """"""
    VOLUMES = "volumes"
    """"""
    PROCESSING_PATH = "processingPath"
    """"""
    VOLUME_ID = "volumeId"
    """"""
    VOLUME_SLICES = "volumeSlices"
    """"""
    VOLUME_IDS = "volumeIds"
    """"""
    VOLUME_NAME = "volumeName"
    """"""
    SIZEB3 = (["fileMeta", "size"], "sizeb")
    """"""
    INCLUDE = "include"
    """"""
    CLASSES = "classes"
    """"""
    PROJECT_TAGS = "projectTags"
    """"""
    DATASETS = "datasets"
    """"""
    IMAGES_TAGS = "imagesTags"
    """"""
    ANNOTATION_OBJECTS_TAGS = "annotationObjectsTags"
    """"""
    FIGURES_TAGS = "figuresTags"
    """"""
    REDIRECT_REQUESTS = "redirectRequests"
    """"""
    PROCESSING_PATH = "processingPath"
    """"""
    FORCE_METADATA_FOR_LINKS = "forceMetadataForLinks"
    """"""
    SKIP_VALIDATION = "skipValidation"
    """"""
    PAGINATION_MODE = "pagination_mode"
    """"""
    PER_PAGE = "per_page"
    """"""
    SKIP_BOUNDS_VALIDATION = "skipBoundsValidation"
    """"""
    PATHS = "paths"
    """"""
    PROJECTS = "projects"
    """"""
    URL = "url"
    """"""
    ANN_URL = "annotationsUrl"
    """"""
    ARCHIVE_URL = "archiveUrl"
    """"""
    ANN_ARCHIVE_URL = "annotationsArchiveUrl"
    """"""
    BACKUP_ARCHIVE = "backupArchive"
    """"""
    SKIP_EXPORTED = "skipExported"
    """"""
    FROM = "from"
    """"""
    TO = "to"
    """"""
    DATA = "data"
    """"""
    DURATION = "duration"
    """"""
    RAW_VIDEO_META = "rawVideoMeta"
    """"""
    IS_DIR = "isDir"
    """"""
    FIGURE_CLASS_ID = "figureClassId"
    """"""
    FIGURE_CLASS_TITLE = "figureClassTitle"
    """"""
    TOOL_CLASS_ID = "toolClassId"
    """"""
    TOOL_STATE = "toolState"
    """"""
    OPTION = "option"
    """"""
    DECOMPRESS_BITMAP = "decompressBitmap"
    """"""
    FIGURE_STATE = "figureState"
    """"""
    BITMAP = "bitmap"
    """"""
    ACCOUNT_TYPE = "plan"
    """"""
    SILENT = "silent"
    """"""
    PAGE = "page"
    """"""
    PRESERVE_PROJECT_CARD = "preserveProjectCard"
    """"""
    GPU_INFO = "gpuInfo"
    """"""
    IS_PUBLIC = "isPublic"
    """"""
    USAGE = "usage"
    """"""
    PARENT_ID = "parentId"
    """"""
    SRC_ID = "srcId"
    """"""
    DEST_ID = "destId"
    """"""
    COMPLETE_EXISTING = "completeExisting"
    """"""
    MODE = "mode"
    """"""
    FOLDERS = "folders"
    """"""
    WITH_METADATA = "withMetadata"
    """"""
    LIMIT = "limit"
    """"""
    IMPORT_SETTINGS = "importSettings"
    """"""
    ADVANCED_SETTINGS = "advancedSettings"
    """"""
    LIMIT_BY_WORKSPACE = "limitByWorkspace"
    """"""
    COMMIT_TOKEN = "commitToken"
    """"""
    TEAM_FILE_ID = "teamFileId"
    """"""
    PROJECT_UPDATED_AT = "projectUpdatedAt"
    """"""
    NODE = "node"
    """"""
    OPERATOR = "operator"
    """"""
    SMART_TOOL_INPUT = "smartToolInput"
    """"""
    LABELING_QUEUE_ID = "labelingQueueId"
    """"""
    LABELING_EXAM_ID = "labelingExamId"
    """"""
    IS_PINNED = "isPinned"
    """"""
    EVENTS = "events"
    """"""
    COMMENTS = "comments"
    """"""
    COMMENT = "comment"
    """"""
    ASSIGNEES = "assignees"
    """"""
    LINKS = "links"
    """"""
    IS_LOCAL = "isLocal"
    """"""
    ISSUE_ID = "issueId"
    """"""
    EXTRA = "extra"
    """"""
    FIGURE_IMAGE_ID = "figureImageId"
    """"""
    BINDINGS = "bindings"
    """"""
    TOP = "top"
    """"""
    LEFT = "left"
    """"""
    POSITION = "position"
    """"""
    ANNOTATION_DATA = "annotationData"
    """"""
    JOB_ENTITY_STATUS = "jobEntityStatus"
    """"""
    INTEGER_COORDS = "integerCoords"
    """"""
    HIDDEN = "hidden"
    """"""
    PRIORITY = "priority"
    """"""
    CLEAR_LOCAL_DATA_SOURCE = "clearLocalDataSource"
    """"""
    ONLY_RUNNING = "onlyRunning"
    """"""
    SHOW_DISABLED = "showDisabled"
    """"""
    WITH_SHARED = "withShared"
    """"""
    USE_DIRECT_PROGRESS_MESSAGES = "useDirectProgressMessages"
    """"""
    EXTRA_FIELDS = "extraFields"
    """"""
    CUSTOM_SORT = "customSort"
    """"""
    GROUP_ID = "groupId"
    """"""
    EXPERIMENT = "experiment"
    """"""
    IS_FINISHED = "isFinished"
    """"""
    NON_FINAL_VALUE = "nonFinalValue"
    """"""
    HOTKEY = "hotkey"
    """"""
    RELATED_DATA_ID = "relatedDataId"
    """"""
    DOWNLOAD_ID = "downloadId"
    """"""
    OFFSET_START = "offsetStart"
    """"""
    OFFSET_END = "offsetEnd"
    """"""
    SOURCE_BLOB = "sourceBlob"
    """"""
    JOBS = "jobs"
    """"""
    LABELERS = "labelers"
    """"""
    REVIEWERS = "reviewers"
    """"""
    REVIEWER_IDS = "reviewerIds"
    """"""
    ENTITIES_COUNT = "entitiesCount"
    """"""
    ACCEPTED_COUNT = "acceptedCount"
    """"""
    ANNOTATED_COUNT = "annotatedCount"
    """"""
    IN_PROGRESS_COUNT = "inProgressCount"
    """"""
    PENDING_COUNT = "pendingCount"
    """"""
    QUEUE_META = "queueMeta"
    """"""
    ENTITY_IDS = "entityIds"
    """"""
    COLLECTION_ID = "collectionId"
    """"""
    QUALITY_CHECK_USER_IDS = "qualityCheckUserIds"
    """"""
    EMBEDDINGS = "embeddings"
    """"""
    EMBEDDINGS_ENABLED = "embeddingsEnabled"
    """"""
    EMBEDDINGS_UPDATED_AT = "embeddingsUpdatedAt"
    """"""
    EMBEDDINGS_IN_PROGRESS = "embeddingsInProgress"
    """"""
    AI_SEARCH_KEY = "aiSearchKey"
    """"""
    AI_SEARCH_META = "aiSearchMeta"
    """"""
    ENTITY_ITEMS = "entityItems"
    """"""
    SCORE = "score"
    """"""
    HARD_DELETE = "hardDelete"
    """"""
    THRESHOLD = "threshold"
    """"""
    THRESHOLD_DIRECTION = "thresholdDirection"
    """"""
    METHOD = "method"
    """"""
    PROMPT = "prompt"
    """"""
    UPDATE_STRATEGY = "updateStrategy"
    """"""
    FILE_PATH = "filePath"
    """"""
    FILE_KEY = "fileKey"
    """"""
    LOCAL_ENTITIES_COUNT = "localEntitiesCount"
    """"""
    REMOTE_ENTITIES_COUNT = "remoteEntitiesCount"
    """"""
    RESTRICTED_IMAGE_IDS = "restrictedImageIds"
    """"""
    NUMBER_OF_CLUSTERS = "numberOfClusters"
    """"""
    CLUSTERING_METHOD = "clusteringMethod"
    """"""
    ERROR_MESSAGE = "errorMessage"
    """"""
    UNIQUE_ITEMS = "uniqueItems"
    """"""
    NN_CREATED = "nnCreated"
    """"""
    NN_UPDATED = "nnUpdated"
    """"""
    M_GUIDE_ID = (["meta", "guide"], "guide_id")
    """"""
    GUIDE_ID = "guideId"
    """"""
    SINGLE_SESSION_MODE = "singleSessionMode"


def _get_single_item(items):
    """_get_single_item"""
    if len(items) == 0:
        return None
    if len(items) > 1:
        raise RuntimeError("There are several items with the same name")
    return items[0]


class _JsonConvertibleModule:
    """Base class for all module APIs that can be converted to JSON."""

    def _convert_json_info(self, info: dict, skip_missing=False):
        """
        Convert JSON info to a dictionary.

        :param info: Dictionary with JSON info.
        :type info: dict
        :param skip_missing: Skip missing fields.
        :type skip_missing: bool
        :returns: Dictionary with JSON info.
        :rtype: dict
        """
        raise NotImplementedError()


class ModuleApiBase(_JsonConvertibleModule):
    """Base class for all module APIs."""

    MAX_WAIT_ATTEMPTS = 999
    """ Maximum number of attempts that will be made to wait for a certain condition to be met."""
    WAIT_ATTEMPT_TIMEOUT_SEC = 1
    """Number of seconds for intervals between attempts."""

    @staticmethod
    def info_sequence():
        """Get list of all class field names."""

        raise NotImplementedError()

    @staticmethod
    def info_tuple_name():
        """Get string name of NamedTuple."""
        raise NotImplementedError()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        try:
            field_names = []
            for name in cls.info_sequence():
                if type(name) is str:
                    field_names.append(camel_to_snake(name))
                elif type(name) is tuple and type(name[1]) is str:
                    field_names.append(name[1])
                else:
                    raise RuntimeError("Can not parse field {!r}".format(name))
            cls.InfoType = namedtuple(cls.info_tuple_name(), field_names)
        except NotImplementedError:
            pass

    def __init__(self, api: "Api"):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        self._api = api

    def _add_sort_param(self, data):
        """
        Add sort parameter to the data.

        :param data: Dictionary with request body info.
        :type data: dict
        :returns: Dictionary with request body info and sort parameter.
        :rtype: dict
        """
        results = deepcopy(data)
        results[ApiField.SORT] = ApiField.ID
        results[ApiField.SORT_ORDER] = "asc"  # @TODO: move to enum
        return results

    def get_list_all_pages(
        self,
        method,
        data,
        progress_cb=None,
        convert_json_info_cb=None,
        limit: int = None,
        return_first_response: bool = False,
    ):
        """
        Get list of all or limited quantity entities from the Supervisely server.

        :param method: Request method name
        :type method: str
        :param data: Dictionary with request body info
        :type data: dict
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :param convert_json_info_cb: Function for convert json info
        :type convert_json_info_cb: Callable, optional
        :param limit: Number of entity to retrieve
        :type limit: int, optional
        :param return_first_response: Specify if return first response
        :type return_first_response: bool, optional
        :returns: List of entities.
        :rtype: List[dict]
        """

        if convert_json_info_cb is None:
            convert_func = self._convert_json_info
        else:
            convert_func = convert_json_info_cb

        if ApiField.SORT not in data:
            data = self._add_sort_param(data)
        first_response = self._api.post(method, data).json()
        total = first_response["total"]
        per_page = first_response["perPage"]
        pages_count = first_response["pagesCount"]

        limit_exceeded = False
        results = first_response["entities"]
        if limit is not None and len(results) > limit:
            limit_exceeded = True

        if progress_cb is not None:
            progress_cb(len(results))
        if (pages_count == 1 and len(results) == total) or limit_exceeded is True:
            pass
        else:
            for page_idx in range(2, pages_count + 1):
                temp_resp = self._api.post(method, {**data, "page": page_idx, "per_page": per_page})
                temp_items = temp_resp.json()["entities"]
                results.extend(temp_items)
                if progress_cb is not None:
                    progress_cb(len(temp_items))
                if limit is not None and len(results) > limit:
                    limit_exceeded = True
                    break

            if len(results) != total and limit is None:
                raise RuntimeError(
                    "Method {!r}: error during pagination, some items are missed".format(method)
                )

        if limit is not None:
            results = results[:limit]
        if return_first_response:
            return [convert_func(item) for item in results], first_response
        return [convert_func(item) for item in results]

    def get_list_all_pages_generator(
        self,
        method,
        data,
        progress_cb=None,
        convert_json_info_cb=None,
        limit: int = None,
        return_first_response: bool = False,
    ):
        """
        This generator function retrieves a list of all or a limited quantity of entities from the Supervisely server, yielding batches of entities as they are retrieved

        :param method: Request method name
        :type method: str
        :param data: Dictionary with request body info
        :type data: dict
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: Progress, optional
        :param convert_json_info_cb: Function for convert json info
        :type convert_json_info_cb: Callable, optional
        :param limit: Number of entity to retrieve
        :type limit: int, optional
        :param return_first_response: Specify if return first response
        :type return_first_response: bool, optional
        """

        if convert_json_info_cb is None:
            convert_func = self._convert_json_info
        else:
            convert_func = convert_json_info_cb

        if ApiField.SORT not in data:
            data = self._add_sort_param(data)
        first_response = self._api.post(method, data).json()
        total = first_response["total"]
        # per_page = first_response["perPage"]
        after = first_response["after"]
        # pages_count = first_response["pagesCount"]

        limit_exceeded = False
        results = first_response["entities"]
        processed = len(results)
        yield [convert_func(item) for item in results]
        if limit is not None and len(results) > limit:
            limit_exceeded = True

        if progress_cb is not None:
            progress_cb(len(results))
        if len(first_response["entities"]) == total or limit_exceeded is True:
            pass
        else:
            while after is not None:
                temp_resp = self._api.post(method, {**data, "after": after}).json()
                after = temp_resp.get("after")
                results = temp_resp["entities"]
                # results.extend(temp_items)
                if progress_cb is not None:
                    progress_cb(len(results))
                processed += len(results)
                yield [convert_func(item) for item in results]
                if limit is not None and processed > limit:
                    limit_exceeded = True
                    break

            if processed != total and limit is None:
                raise RuntimeError(
                    "Method {!r}: error during pagination, some items are missed".format(method)
                )

    @staticmethod
    def _get_info_by_name(get_info_by_filters_fn, name):
        """
        Get information about an entity by its name from the Supervisely server.

        :param get_info_by_filters_fn: Function to get information about an entity by its filters.
        :type get_info_by_filters_fn: Callable
        :param name: Name of the entity.
        :type name: str
        :returns: Information about the entity.
        :rtype: dict
        """
        filters = [{"field": ApiField.NAME, "operator": "=", "value": name}]
        return get_info_by_filters_fn(filters)

    def get_info_by_id(self, id):
        """
        Get information about an entity by its ID from the Supervisely server.

        :param id: ID of the entity.
        :type id: int
        """

        raise NotImplementedError()

    @staticmethod
    def _get_free_name(exist_check_fn, name):
        """
        Get a free name for an entity.

        :param exist_check_fn: Function to check if a name is already used.
        :type exist_check_fn: Callable
        :param name: Name of the entity.
        :type name: str
        :returns: Free name for the entity.
        :rtype: str
        """
        res_title = name
        suffix = 1
        while exist_check_fn(res_title):
            res_title = "{}_{:03d}".format(name, suffix)
            suffix += 1
        return res_title

    def _convert_json_info(self, info: dict, skip_missing=False):
        """
        Convert JSON info to a dictionary.

        :param info: Dictionary with JSON info.
        :type info: dict
        :param skip_missing: Skip missing fields.
        :type skip_missing: bool
        :returns: Dictionary with JSON info.
        :rtype: dict
        """

        def _get_value(dict, field_name, skip_missing):
            if skip_missing is True:
                return dict.get(field_name, None)
            else:
                return dict[field_name]

        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if type(field_name) is str:
                    field_values.append(_get_value(info, field_name, skip_missing))
                elif type(field_name) is tuple:
                    value = None
                    for sub_name in field_name[0]:
                        if value is None:
                            value = _get_value(info, sub_name, skip_missing)
                        else:
                            value = _get_value(value, sub_name, skip_missing)
                    field_values.append(value)
                else:
                    raise RuntimeError("Can not parse field {!r}".format(field_name))
            return self.InfoType(*field_values)

    @classmethod
    def convert_info_to_json(cls, info: NamedTuple) -> Dict:
        """
        Convert information about an entity to a dictionary.

        :param info: Information about the entity.
        :type info: NamedTuple
        :returns: Dictionary with information about the entity.
        :rtype: dict
        """

        def _create_nested_dict(keys, value):
            if len(keys) == 1:
                return {keys[0]: value}
            else:
                return {keys[0]: _create_nested_dict(keys[1:], value)}

        json_info = {}
        for field_name, value in zip(cls.info_sequence(), info):
            if type(field_name) is str:
                json_info[field_name] = value
            elif isinstance(field_name, tuple):
                if len(field_name[0]) == 0:
                    json_info[field_name[1]] = value
                else:
                    json_info.update(_create_nested_dict(field_name[0], value))
            else:
                raise RuntimeError("Can not parse field {!r}".format(field_name))
        return json_info

    def _get_response_by_id(self, id, method, id_field, fields=None):
        """_get_response_by_id"""
        try:
            data = {id_field: id}
            if fields is not None:
                data.update(fields)
            return self._api.post(method, data)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                return None
            else:
                raise error

    def _get_info_by_id(self, id, method, fields=None):
        """_get_info_by_id"""
        response = self._get_response_by_id(id, method, id_field=ApiField.ID, fields=fields)
        return self._convert_json_info(response.json()) if (response is not None) else None

    async def get_list_idx_page_async(
        self,
        method: str,
        data: dict,
    ) -> Tuple[int, List[NamedTuple]]:
        """
        Get the list of items for a given page number.
        Page number is specified in the data dictionary.

        :param method: Method to call for listing items.
        :type method: str
        :param data: Data to pass to the API method.
        :type data: dict
        :returns: List of items.
        :rtype: Tuple[int, List[NamedTuple]]
        """

        response = await self._api.post_async(method, data)
        response_json = response.json()
        entities = response_json.get("entities", [])
        # To avoid empty pages when a filter is applied to the data and the `pagesCount` is less than the number calculated based on the items and `per_page` size.
        # Process `pagesCount` in the main function according to the actual number of pages returned.
        pages_count = response_json.get("pagesCount", None)
        if pages_count is None:
            raise ValueError("Can not determine the number of pages to retrieve.")
        return pages_count, [self._convert_json_info(item) for item in entities]

    async def get_list_page_generator_async(
        self,
        method: str,
        data: dict,
        pages_count: Optional[int] = None,
        semaphore: Optional[List[asyncio.Semaphore]] = None,
    ) -> AsyncGenerator[List[Any], None]:
        """
        Yields list of images in dataset asynchronously page by page.

        :param method: Method to call for listing items.
        :type method: str
        :param data: Data to pass to the API method.
        :type data: dict
        :param pages_count: Preferred number of pages to retrieve if used with a ``per_page`` limit.
                            Will be automatically adjusted if the ``pagesCount`` differs from the requested number.
        :type pages_count: int, optional
        :param semaphore: Semaphore for limiting the number of simultaneous requests.
        :type semaphore: :class:`asyncio.Semaphore`, optional
        :returns: List of images in dataset.
        :rtype: AsyncGenerator[List[:class:`~supervisely.api.image_api.ImageInfo`]]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                method = 'images.list'
                data = {'datasetId': 123456}

                loop = sly.utils.get_or_create_event_loop()
                images = loop.run_until_complete(api.image.get_list_generator_async(method, data))
        """

        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def sem_task(task):
            async with semaphore:
                return await task

        if pages_count is None:
            pages_count = 999999  # to avoid range lesser than total pages count
        for page_num in range(1, pages_count + 1):
            if page_num <= pages_count:
                data[ApiField.PAGE] = page_num
                total_pages, items = await sem_task(self.get_list_idx_page_async(method, data))

                # To correct `total_pages` count in case filter is applied
                if page_num == 1:
                    pages_count = total_pages

                yield items
            else:
                break


class ModuleApi(ModuleApiBase):
    """Base class for entities that have a parent object in the system."""

    MAX_WAIT_ATTEMPTS = ModuleApiBase.MAX_WAIT_ATTEMPTS
    """Maximum number of attempts that will be made to wait for a certain condition to be met."""

    WAIT_ATTEMPT_TIMEOUT_SEC = ModuleApiBase.WAIT_ATTEMPT_TIMEOUT_SEC
    """Number of seconds for intervals between attempts."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        super().__init__(api)
        self._api = api

    def get_info_by_name(self, parent_id: int, name: str, fields: List[str] = []):
        """
        Get information about an entity by its name from the Supervisely server.

        :param parent_id: ID of the parent entity.
        :type parent_id: int
        :param name: Name of the entity for which the information is being retrieved.
        :type name: str
        :param fields: The list of api fields which will be returned with the response.
        :type fields: List[str]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 55832
                name = "IMG_0315.jpeg"
                info = api.image.get_info_by_name(dataset_id, name)
                print(info)
                # Output: ImageInfo(id=19369643, name='IMG_0315.jpeg', ...)
        """

        return self._get_info_by_name(
            get_info_by_filters_fn=lambda module_name: self._get_info_by_filters(
                parent_id, module_name, fields
            ),
            name=name,
        )

    def _get_info_by_filters(self, parent_id, filters, fields=[]):
        """_get_info_by_filters"""
        # pylint: disable=too-many-function-args
        from supervisely.api.project_api import ProjectApi

        if type(self) == ProjectApi:
            items = self.get_list(parent_id, filters, fields)
        else:
            items = self.get_list(parent_id, filters)
        return _get_single_item(items)

    def get_list(self, parent_id, filters=None):
        """
        Get list of entities in parent entity with given parent ID.

        :param parent_id: parent ID in Supervisely.
        :type parent_id: int
        :param filters: List of parameters to sort output entities.
        :type filters: List[Dict[str, str]], optional

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                dataset_id = 55832
                images = api.image.get_list(dataset_id)
                print(images)
                # Output: [
                    ImageInfo(id=19369642, ...)
                    ImageInfo(id=19369643, ...)
                    ImageInfo(id=19369644, ...)
                ]
        """

        raise NotImplementedError()

    def exists(self, parent_id, name):
        """
        Checks if an entity with the given parent_id and name exists

        :param parent_id: ID of the parent entity.
        :type parent_id: int
        :param name: Name of the entity.
        :type name: str
        :returns: Returns True if entity exists, and False if not
        :rtype: bool

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                name = "IMG_0315.jpeg"
                dataset_id = 55832
                exists = api.image.exists(dataset_id, name)
                print(exists) # True
        """

        return self.get_info_by_name(parent_id, name) is not None

    def get_free_name(self, parent_id, name):
        """
        Generates a free name for an entity with the given parent_id and name.
        Adds an increasing suffix to original name until a unique name is found.

        :param parent_id: ID of the parent entity.
        :type parent_id: int
        :param name: Name of the entity.
        :type name: str
        :returns: Returns free name.
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                name = "IMG_0315.jpeg"
                dataset_id = 55832
                free_name = api.image.get_free_name(dataset_id, name)
                print(free_name) # IMG_0315_001.jpeg
        """

        return self._get_free_name(
            exist_check_fn=lambda module_name: self.exists(parent_id, module_name),
            name=name,
        )

    def _get_effective_new_name(self, parent_id, name, change_name_if_conflict=False):
        """_get_effective_new_name"""
        return self.get_free_name(parent_id, name) if change_name_if_conflict else name


# Base class for entities that do not have a parent object in the system.
class ModuleNoParent(ModuleApiBase):
    """Base API module for entities that are not nested under a parent object (parent_id not required)."""

    def get_info_by_name(self, name):
        """get_info_by_name"""
        return self._get_info_by_name(get_info_by_filters_fn=self._get_info_by_filters, name=name)

    def _get_info_by_filters(self, filters):
        """_get_info_by_filters"""
        items = self.get_list(filters)
        return _get_single_item(items)

    def get_list(self, filters=None):
        """get_list"""
        raise NotImplementedError()

    def exists(self, name):
        """exists"""
        return self.get_info_by_name(name) is not None

    def get_free_name(self, name):
        """get_free_name"""
        return self._get_free_name(
            exist_check_fn=lambda module_name: self.exists(module_name), name=name
        )

    def _get_effective_new_name(self, name, change_name_if_conflict=False):
        """"""
        return self.get_free_name(name) if change_name_if_conflict else name


class CloneableModuleApi(ModuleApi):
    """API for cloning entities."""

    MAX_WAIT_ATTEMPTS = ModuleApiBase.MAX_WAIT_ATTEMPTS
    """Maximum number of attempts that will be made to wait for a certain condition to be met."""

    WAIT_ATTEMPT_TIMEOUT_SEC = ModuleApiBase.WAIT_ATTEMPT_TIMEOUT_SEC
    """Number of seconds for intervals between attempts."""

    def _clone_api_method_name(self):
        """_clone_api_method_name"""
        raise NotImplementedError()

    def _clone(self, clone_type: dict, dst_workspace_id: int, dst_name: str):
        """_clone"""
        response = self._api.post(
            self._clone_api_method_name(),
            {
                **clone_type,
                ApiField.WORKSPACE_ID: dst_workspace_id,
                ApiField.NAME: dst_name,
            },
        )
        return response.json()[ApiField.TASK_ID]

    def clone(self, id: int, dst_workspace_id: int, dst_name: str) -> int:
        """Clones the entity with the given ID to the given workspace with the given name.
        Returns the ID of the task that is created to perform the cloning operation.

        :param id: ID of the entity to clone.
        :type id: int
        :param dst_workspace_id: ID of the workspace to clone to.
        :type dst_workspace_id: int
        :param dst_name: Name of the cloned entity.
        :type dst_name: str
        :returns: Returns the ID of the task that is created to perform the cloning operation.
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                # ID of the entity to clone
                project_id = 123456

                # ID of the destination workspace
                workspace_id = 123456

                # Create a task to clone the project
                task_id = api.project.clone(project_id, workspace_id, "my_cloned_project")

                # Wait until the task is finished
                api.task.wait(task_id, api.task.Status.FINISHED)

                task_info = api.task.get_info_by_id(task_id)

                dst_project_id = task_info["meta"]["output"]["project"]["id"]
                print(f"Cloned project ID: {dst_project_id}")
        """

        return self._clone({ApiField.ID: id}, dst_workspace_id, dst_name)

    def clone_by_shared_link(self, shared_link: str, dst_workspace_id: int, dst_name: str) -> int:
        """Clones the entity with the given shared link to the given workspace with the given name.
        Returns the ID of the task that is created to perform the cloning operation.

        :param shared_link: Shared link of the entity to clone.
        :type shared_link: str
        :param dst_workspace_id: ID of the workspace to clone to.
        :type dst_workspace_id: int
        :param dst_name: Name of the cloned entity.
        :type dst_name: str
        :returns: Returns the ID of the task that is created to perform the cloning operation.
        :rtype: int
        """
        return self._clone({ApiField.SHARED_LINK: shared_link}, dst_workspace_id, dst_name)

    def clone_from_explore(self, explore_path: str, dst_workspace_id: int, dst_name: str) -> int:
        """Clones the entity with the given explore path to the given workspace with the given name.
        Returns the ID of the task that is created to perform the cloning operation.

        :param explore_path: Explore path of the entity to clone.
        :type explore_path: str
        :param dst_workspace_id: ID of the workspace to clone to.
        :type dst_workspace_id: int
        :param dst_name: Name of the cloned entity.
        :type dst_name: str
        :returns: Returns the ID of the task that is created to perform the cloning operation.
        :rtype: int
        """
        return self._clone({ApiField.EXPLORE_PATH: explore_path}, dst_workspace_id, dst_name)

    def get_or_clone_from_explore(self, explore_path, dst_workspace_id, dst_name):
        """get_or_clone_from_explore"""
        if not self.exists(dst_workspace_id, dst_name):
            task_id = self.clone_from_explore(explore_path, dst_workspace_id, dst_name)
            self._api.task.wait(task_id, self._api.task.Status.FINISHED)
        item = self.get_info_by_name(dst_workspace_id, dst_name)
        return item


class ModuleWithStatus:
    """API for getting and raising for status of an entity."""

    def get_status(self, id):
        """get_status"""
        raise NotImplementedError()

    def raise_for_status(self, status):
        """raise_for_status"""
        raise NotImplementedError()


class WaitingTimeExceeded(Exception):
    """Exception raised when waiting for a status of an entity exceeds the timeout."""

    pass


class UpdateableModule(_JsonConvertibleModule):
    """API for updating entities."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        self._api = api

    def _get_update_method(self):
        """Get the method name for updating an entity."""
        raise NotImplementedError()

    def update(self, id, name=None, description=None):
        """Update an entity with the specified ID.

        :param id: ID of the entity to update.
        :type id: int
        :param name: New name of the entity.
        :type name: str, optional
        :param description: New description of the entity.
        :type description: str, optional
        :returns: Entity with updated information.
        :rtype: dict
        """
        if name is None and description is None:
            raise ValueError("'name' or 'description' or both have to be specified")

        body = {ApiField.ID: id}
        if name is not None:
            body[ApiField.NAME] = name
        if description is not None:
            body[ApiField.DESCRIPTION] = description

        response = self._api.post(self._get_update_method(), body)
        return self._convert_json_info(response.json())


class RemoveableModuleApi(ModuleApi):
    """API for removing entities."""

    MAX_WAIT_ATTEMPTS = ModuleApiBase.MAX_WAIT_ATTEMPTS
    """Maximum number of attempts that will be made to wait for a certain condition to be met."""

    WAIT_ATTEMPT_TIMEOUT_SEC = ModuleApiBase.WAIT_ATTEMPT_TIMEOUT_SEC
    """Number of seconds for intervals between attempts."""

    def _remove_api_method_name(self):
        """Get the method name for removing an entity."""
        raise NotImplementedError()

    def remove(self, id):
        """
        Remove an entity with the specified ID from the Supervisely server.

        :param id: Entity ID in Supervisely
        :type id: int
        :returns: None
        :rtype: None
        """
        self._api.post(self._remove_api_method_name(), {ApiField.ID: id})

    def remove_batch(self, ids, progress_cb=None):
        """
        Remove entities with given IDs from the Supervisely server.

        :param ids: IDs of entities in Supervisely.
        :type ids: List[int]
        :param progress_cb: Function for control remove progress.
        :type progress_cb: Callable
        :returns: None
        :rtype: None
        """
        for id in ids:
            self.remove(id)
            if progress_cb is not None:
                progress_cb(1)


class RemoveableBulkModuleApi(ModuleApi):
    """API for removing multiple entities."""

    MAX_WAIT_ATTEMPTS = ModuleApiBase.MAX_WAIT_ATTEMPTS
    """Maximum number of attempts that will be made to wait for a certain condition to be met."""

    WAIT_ATTEMPT_TIMEOUT_SEC = ModuleApiBase.WAIT_ATTEMPT_TIMEOUT_SEC
    """Number of seconds for intervals between attempts."""

    def _remove_batch_api_method_name(self):
        """Get the method name for removing a batch of entities."""
        raise NotImplementedError()

    def _remove_batch_field_name(self):
        """Get the field name for the IDs of the entities to remove."""
        raise NotImplementedError()

    def remove_batch(self, ids, progress_cb=None, batch_size=50):
        """
        Remove entities in batches from the Supervisely server.
        All entity IDs must belong to the same nesting (for example team, or workspace, or project, or dataset).
        Therefore, it is necessary to sort IDs before calling this method.

        :param ids: IDs of entities in Supervisely.
        :type ids: List[int]
        :param progress_cb: Function for control remove progress.
        :type progress_cb: Callable

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                image_ids = [19369645, 19369646, 19369647]
                api.image.remove_batch(image_ids)
        """
        for ids_batch in batched(ids, batch_size=batch_size):
            self._api.post(
                self._remove_batch_api_method_name(),
                {self._remove_batch_field_name(): ids_batch},
            )
            if progress_cb is not None:
                progress_cb(len(ids_batch))

    def remove(self, id):
        """
        Remove an entity with the specified ID from the Supervisely server.

        :param id: Entity ID in Supervisely.
        :type id: int

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                image_id = 19369643
                api.image.remove(image_id)
        """
        self.remove_batch([id])



from typing import Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class InputNumber(Widget):
    """Numeric input widget with min/max/step/precision and value-changed callback."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""

        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        value: Union[int, float] = 1,
        min: Union[int, float, None] = None,
        max: Union[int, float, None] = None,
        step: Union[int, float] = 1,
        size: str = "small",
        controls: bool = True,
        debounce: int = 300,
        precision: int = 0,
        widget_id: str = None,
        width: int = None,
    ):
        """
        :param value: Initial numeric value.
        :type value: Union[int, float]
        :param min: Minimum value. None for no limit.
        :type min: Union[int, float, None], optional
        :param max: Maximum value. None for no limit.
        :type max: Union[int, float, None], optional
        :param step: Step increment.
        :type step: Union[int, float]
        :param size: Size ("small", "large", etc.).
        :type size: str
        :param controls: If True, show +/- buttons.
        :type controls: bool
        :param debounce: Debounce delay in ms for value-changed callback.
        :type debounce: int
        :param precision: Decimal places for float values.
        :type precision: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param width: Width in pixels. None for auto.
        :type width: int, optional
        """
        self._value = value
        self._min = min
        self._max = max
        self._step = step
        self._size = size
        self._controls = controls
        self._debounce = debounce
        self._precision = precision
        self._width = width
        self._changes_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "min": self._min,
            "max": self._max,
            "step": self._step,
            "size": self._size,
            "controls": self._controls,
            "debounce": self._debounce,
            "precision": self._precision,
        }

    def get_json_state(self):
        return {"value": self._value}

    @property
    def value(self):
        self._value = StateJson()[self.widget_id]["value"]
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self):
        self._value = StateJson()[self.widget_id]["value"]
        return self._value

    @property
    def min(self):
        return self._min

    @min.setter
    def min(self, value):
        self._min = value
        DataJson()[self.widget_id]["min"] = self._min
        DataJson().send_changes()

    @property
    def max(self):
        return self._max

    @max.setter
    def max(self, value):
        self._max = value
        DataJson()[self.widget_id]["max"] = self._max
        DataJson().send_changes()

    def value_changed(self, func):
        route_path = self.get_route_path(InputNumber.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            res = self.get_value()
            self._value = res
            func(res)

        return _click



# coding: utf-8

from supervisely.api.module_api import ModuleApiBase, ApiField


class ImportStorageApi(ModuleApiBase):
    """Work with internal import storage in Supervisely."""

    def get_meta_by_hashes(self, hashes):
        """
        """
        response = self._api.post('import-storage.internal.meta.list', {ApiField.HASHES: hashes})
        return response.json()



import os
from collections import defaultdict
from pathlib import Path

from supervisely import Api, ProjectMeta, logger
from supervisely._utils import batched, is_development
from supervisely.annotation.obj_class import ObjClass
from supervisely.convert.volume.nii import nii_volume_helper as helper
from supervisely.convert.volume.nii.nii_volume_converter import NiiConverter
from supervisely.convert.volume.volume_converter import VolumeConverter
from supervisely.geometry.mask_3d import Mask3D
from supervisely.io.fs import get_file_ext, get_file_name, list_files_recursively
from supervisely.volume.volume import is_nifti_file
from supervisely.volume_annotation.volume_annotation import VolumeAnnotation
from supervisely.volume_annotation.volume_object import VolumeObject


class NiiPlaneStructuredConverter(NiiConverter, VolumeConverter):
    """Convert NIfTI 3D volume file to Supervisely format.
    The NIfTI file should be structured as follows:
    - <prefix>_anatomic_<idx>.nii (or .nii.gz)
    - <prefix>_inference_<idx>.nii (or .nii.gz)
        where   <prefix> is one of the following: cor, sag, axl
                <idx> is the index of the volume (to match volumes with annotations)

    Supports .nii and .nii.gz files.

    Example:
        📂 .
        ├── 🩻 axl_anatomic_1.nii
        ├── 🩻 axl_inference_1.nii class 1 (may contain multiple instances of the same class)
        ├── 🩻 cor_anatomic_1.nii
        ├── 🩻 cor_inference_1.nii class 1
        ├── 🩻 sag_anatomic_1.nii
        ├── 🩻 sag_inference_1.nii class 1
        ├── 🩻 sag_inference_2.nii class 2
        └── 🩻 sag_inference_3.nii class 3

    Additionally, if a TXT file with class color map is present, it will be used to
    create the classes with names and colors corresponding to the pixel values in the NIfTI files.
    The TXT file should be structured as follows:

    ```txt
    1 Femur 255 0 0
    2 Femoral cartilage 0 255 0
    3 Tibia 0 0 255
    4 Tibia cartilage 255 255 0
    5 Patella 0 255 255
    6 Patellar cartilage 255 0 255
    7 Miniscus 175 175 175
    ```
    where   1, 2, ... are the pixel values in the NIfTI files
            Femur, Femoral cartilage, ... are the names of the classes
            255, 0, 0, ... are the RGB colors of the classes
    The class name will be used to create the corresponding ObjClass in Supervisely.
    """

    class Item(VolumeConverter.BaseItem):
        """Volume item extended with NIfTI-specific flags and a prepared ``volume_meta`` for annotations."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._is_semantic = False
            self.volume_meta = None

        @property
        def is_semantic(self) -> bool:
            return self._is_semantic

        @is_semantic.setter
        def is_semantic(self, value: bool):
            self._is_semantic = value

        def create_empty_annotation(self):
            return VolumeAnnotation(self.volume_meta)

    def __str__(self):
        return "nii_custom"

    def validate_format(self) -> bool:
        # create Items
        converted_dir_name = "converted"

        volumes_dict = defaultdict(list)
        ann_dict = defaultdict(list)
        cls_color_map = None

        ann_to_score_path = {}
        csv_files = list_files_recursively(self._input_data, [".csv"], None, True)
        csv_nameparts = {
            helper.parse_name_parts(os.path.basename(file)): file for file in csv_files
        }

        for root, _, files in os.walk(self._input_data):
            if converted_dir_name in root:
                continue
            for file in files:
                path = os.path.join(root, file)
                if is_nifti_file(path):
                    name_parts = helper.parse_name_parts(file)
                    if name_parts is None:
                        logger.debug(
                            f"File recognized as NIfTI, but failed to parse plane identifier from name. Path: {path}",
                        )
                        continue
                    if name_parts.is_ann:
                        dict_to_use = ann_dict
                        score_path = helper.find_best_name_match(
                            name_parts, list(csv_nameparts.keys())
                        )
                        if score_path is not None:
                            full_score_path = csv_nameparts[score_path]
                            ann_to_score_path[name_parts.full_name] = full_score_path
                    else:
                        dict_to_use = volumes_dict

                    if name_parts.patient_uuid is None and name_parts.case_uuid is None:
                        key = name_parts.plane
                    else:
                        key = f"{name_parts.plane}_{name_parts.patient_uuid}_{name_parts.case_uuid}"
                    dict_to_use[key].append(path)
                ext = get_file_ext(path)
                if ext == ".txt":
                    cls_color_map = helper.read_cls_color_map(path)
                    if cls_color_map is None:
                        logger.warning(f"Failed to read class color map from {path}.")

        self._items = []
        for key, paths in volumes_dict.items():
            if len(paths) == 1:
                item = self.Item(item_path=paths[0])
                name_parts = helper.parse_name_parts(os.path.basename(item.path))
                item.ann_data = ann_dict.get(key, [])

                ann_path = os.path.basename(item.ann_data[0]) if item.ann_data else None
                if ann_path in ann_to_score_path:
                    score_path = ann_to_score_path[ann_path]
                    try:
                        scores = helper.get_scores_from_table(score_path, name_parts.plane)
                        item.custom_data["scores"] = scores
                    except Exception as e:
                        logger.warning(f"Failed to read scores from {score_path}: {e}")
                item.is_semantic = len(item.ann_data) == 1
                if cls_color_map is not None:
                    item.custom_data["cls_color_map"] = cls_color_map
                self._items.append(item)
            elif len(paths) > 1:
                logger.info(
                    f"Found {len(paths)} volumes with key {key}. Will try to match them by directories."
                )
                for path in paths:
                    name_parts = helper.parse_name_parts(os.path.basename(path))
                    item = self.Item(item_path=path)
                    possible_ann_paths = []
                    for ann_path in ann_dict.get(key, []):
                        if Path(ann_path).parent == Path(path).parent:
                            possible_ann_paths.append(ann_path)
                    item.ann_data = possible_ann_paths
                    scores_paths = [
                        ann_to_score_path.get(ann_name, None) for ann_name in possible_ann_paths
                    ]
                    scores_paths = [path for path in scores_paths if path is not None]
                    if scores_paths:
                        try:
                            scores = helper.get_scores_from_table(scores_paths[0], name_parts.plane)
                            item.custom_data["scores"] = scores
                        except Exception as e:
                            logger.warning(f"Failed to read scores from {scores_paths[0]}: {e}")
                    item.is_semantic = len(item.ann_data) == 1
                    if cls_color_map is not None:
                        item.custom_data["cls_color_map"] = cls_color_map
                    self._items.append(item)

        self._meta = ProjectMeta()
        return self.items_count > 0

    def to_supervisely(
        self,
        item: VolumeConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> VolumeAnnotation:
        """Convert to Supervisely format."""

        try:
            objs = []
            spatial_figures = []
            scores = item.custom_data.get("scores", {})
            for idx, ann_path in enumerate(item.ann_data, start=1):
                for mask, pixel_value in helper.get_annotation_from_nii(ann_path):
                    class_id = pixel_value if item.is_semantic else idx
                    class_name = f"Segment_{class_id}"
                    color = None
                    if item.custom_data.get("cls_color_map") is not None:
                        class_info = item.custom_data["cls_color_map"].get(class_id)
                        if class_info is not None:
                            class_name, color = class_info
                    class_name = renamed_classes.get(class_name, class_name)
                    obj_class = meta.get_obj_class(class_name)
                    if obj_class is None:
                        obj_class = ObjClass(
                            class_name,
                            Mask3D,
                            color,
                            description=f"{helper.MASK_PIXEL_VALUE}{pixel_value}",
                        )
                        meta = meta.add_obj_class(obj_class)
                        self._meta_changed = True
                        self._meta = meta
                    obj_scores = scores.get(class_id, {})
                    obj_scores = {k: v for k, v in obj_scores.items()}
                    obj = VolumeObject(obj_class, mask_3d=mask, custom_data=obj_scores)
                    spatial_figures.append(obj.figure)
                    objs.append(obj)
            return VolumeAnnotation(item.volume_meta, objects=objs, spatial_figures=spatial_figures)
        except Exception as e:
            logger.warning(f"Failed to convert {item.path} to Supervisely format: {e}")
            return item.create_empty_annotation()


class NiiPlaneStructuredAnnotationConverter(NiiConverter, VolumeConverter):
    """
    Converter for NIfTI-based 3D mask annotations stored as plane-structured volumes.
    """

    class Item(VolumeConverter.BaseItem):
        """Annotation item for a single plane-structured NIfTI mask (or scores table) with conversion flags."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._is_semantic = False
            self._is_scores = False
            self.volume_meta = None

        @property
        def is_semantic(self) -> bool:
            return self._is_semantic

        @is_semantic.setter
        def is_semantic(self, value: bool):
            self._is_semantic = value

        @property
        def is_scores(self) -> bool:
            return self._is_scores

        @is_scores.setter
        def is_scores(self, value: bool):
            self._is_scores = value

        def create_empty_annotation(self):
            return VolumeAnnotation(self.volume_meta)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._json_map = None

    def __str__(self):
        return "nii_custom_ann"

    def validate_format(self) -> bool:
        try:
            from nibabel import filebasedimages, load
        except ImportError:
            raise ImportError(
                "No module named nibabel. Please make sure that module is installed from pip and try again."
            )
        cls_color_map = None

        has_volumes = lambda x: helper.VOLUME_NAME in x
        if list_files_recursively(self._input_data, filter_fn=has_volumes):
            return False

        txts = list_files_recursively(self._input_data, [".txt"], None, True)
        cls_color_map = next(iter(txts), None)
        if cls_color_map is not None:
            cls_color_map = helper.read_cls_color_map(cls_color_map)

        jsons = list_files_recursively(self._input_data, [".json"], None, True)
        json_map = next(iter(jsons), None)
        if json_map is not None:
            self._json_map = helper.read_json_map(json_map)

        is_nii = lambda x: any(x.endswith(ext) for ext in [".nii", ".nii.gz"])
        for root, _, files in os.walk(self._input_data):
            for file in files:
                path = os.path.join(root, file)
                name_parts = helper.parse_name_parts(file)
                if name_parts is None:
                    continue
                if is_nii(file) or name_parts.type == helper.SCORE_NAME:
                    item = self.Item(item_path=None, ann_data=path)
                    item.custom_data["name_parts"] = name_parts
                    if name_parts.is_ann:
                        try:
                            nii = load(path)
                        except filebasedimages.ImageFileError:
                            logger.warning(f"Failed to load NIfTI file: {path}")
                            continue
                        item.set_shape(nii.shape)
                    elif name_parts.type == helper.SCORE_NAME:
                        item.is_scores = True
                        scores = helper.get_scores_from_table(path, name_parts.plane)
                        item.custom_data["scores"] = scores
                    if cls_color_map is not None:
                        item.custom_data["cls_color_map"] = cls_color_map
                    self._items.append(item)

        obj_classes = None
        if cls_color_map is not None:
            obj_classes = [ObjClass(name, Mask3D, color) for name, color in cls_color_map.values()]

        for item in self._items:
            name_parts = item.custom_data.get("name_parts")
            if item.is_scores:
                continue
            if name_parts.ending_idx is None:
                item.is_semantic = True

        self._meta = ProjectMeta(obj_classes=obj_classes)
        return len(self._items) > 0

    def to_supervisely(
        self,
        item: VolumeConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> VolumeAnnotation:
        """Convert to Supervisely format."""
        try:
            objs = []
            spatial_figures = []
            ann_path = item.ann_data
            ann_idx = item.custom_data["name_parts"].ending_idx or 0
            for mask, pixel_id in helper.get_annotation_from_nii(ann_path):
                class_id = pixel_id if item.is_semantic else ann_idx
                class_name = f"Segment_{class_id}"
                color = None
                if item.custom_data.get("cls_color_map") is not None:
                    class_info = item.custom_data["cls_color_map"].get(class_id)
                    if class_info is not None:
                        class_name, color = class_info
                class_name = renamed_classes.get(class_name, class_name)
                obj_class = meta.get_obj_class(class_name)
                if obj_class is None:
                    obj_class = ObjClass(class_name, Mask3D, color)
                    meta = meta.add_obj_class(obj_class)
                    self._meta_changed = True
                    self._meta = meta
                obj = VolumeObject(obj_class, mask_3d=mask)
                spatial_figures.append(obj.figure)
                objs.append(obj)
            return VolumeAnnotation(item.volume_meta, objects=objs, spatial_figures=spatial_figures)
        except Exception as e:
            logger.warning(f"Failed to convert {item.ann_data} to Supervisely format: {e}")
            return item.create_empty_annotation()

    def upload_dataset(
        self, api: Api, dataset_id: int, batch_size: int = 50, log_progress=True
    ) -> None:
        meta, renamed_classes, _ = self.merge_metas_with_conflicts(api, dataset_id)

        matcher = helper.AnnotationMatcher(self._items, dataset_id)
        if self._json_map is not None:
            try:
                matched_dict = matcher.match_from_json(api, self._json_map)
            except Exception as e:
                logger.error(f"Failed to match annotations from a json map: {e}")
                matched_dict = {}
        else:
            matcher.get_volumes(api)
            matched_dict = matcher.match_items()
            if len(matched_dict) != len(self._items):
                extra = {
                    "items count": len(self._items),
                    "matched count": len(matched_dict),
                    "unmatched count": len(self._items) - len(matched_dict),
                }
                logger.warning(
                    "Not all items were matched with volumes. Some items may be skipped.",
                    extra=extra,
                )
        if len(matched_dict) == 0:
            raise RuntimeError(
                "No items were matched with volumes. Please check the input data and try again."
            )

        if log_progress:
            progress, progress_cb = self.get_progress(
                len(matched_dict), "Uploading volumes annotations..."
            )
        else:
            progress_cb = None

        volumeids_to_objects = defaultdict(list)

        for item, volume in sorted(matched_dict.items(), key=lambda pair: pair[0].is_scores):
            item.volume_meta = volume.meta
            if not item.is_scores:
                ann = self.to_supervisely(item, meta, renamed_classes, None)
                if self._meta_changed:
                    meta, renamed_classes, _ = self.merge_metas_with_conflicts(api, dataset_id)
                    self._meta_changed = False
                api.volume.annotation.append(volume.id, ann, volume_info=volume)
            else:
                class_id_to_pixel_value = helper.get_class_id_to_pixel_value_map(meta)
                scores = item.custom_data.get("scores", {})
                if not scores:
                    logger.warning(f"No scores found for {item.ann_data}. Skipping.")
                    continue

                if volume.dataset_id not in volumeids_to_objects:
                    for obj in api.volume.object.get_list(volume.dataset_id):
                        volumeids_to_objects[obj.entity_id].append(obj)

                obj_id_to_class_id = {
                    obj.id: obj.class_id for obj in volumeids_to_objects[volume.id]
                }
                if not obj_id_to_class_id:
                    logger.warning(
                        f"No objects found for volume {volume.id}. Skipping figure updates."
                    )
                    continue

                volume_figure_dict = api.volume.figure.download(
                    volume.dataset_id, [volume.id], skip_geometry=True
                )
                figures_list = volume_figure_dict.get(volume.id, [])
                for figure in figures_list:
                    class_id = obj_id_to_class_id.get(figure.object_id, None)
                    if class_id is None:
                        logger.warning(
                            f"Class ID for figure (id: {figure.id}) not found in volume objects. Skipping figure update.",
                            extra={
                                "obj_id_to_class_id": obj_id_to_class_id,
                                "object_id": figure.object_id,
                            },
                        )
                        continue
                    pixel_value = class_id_to_pixel_value.get(class_id, None)
                    if pixel_value is None:
                        logger.warning(
                            f"Pixel value for class ID {class_id} not found in meta. Skipping figure update."
                        )
                        continue
                    figure_custom_data = scores.get(pixel_value, {})
                    if figure_custom_data:
                        api.volume.figure.update_custom_data(figure.id, figure_custom_data)
                        logger.debug(
                            f"Updated figure {figure.id} with custom data: {figure_custom_data}"
                        )
            progress_cb(1) if log_progress else None

        if log_progress:
            if is_development():
                progress.close()
            logger.info(f"Successfully uploaded {len(matched_dict)} annotations.")



from typing import Union
import supervisely as sly
from supervisely.project.project_meta import ProjectMeta
from supervisely.api.api import Api
from supervisely.annotation.annotation import Annotation


class Gallery:
    """v1 widget for displaying a grid of images with annotations, zoom, and sync options."""

    def __init__(self, task_id, api: Api, v_model, project_meta: ProjectMeta, col_number: int, preview_info=False,
                 enable_zoom=False, resize_on_zoom=False, sync_views=False, show_preview=True, selectable=False,
                 opacity=0.5, show_opacity_header=True, fill_rectangle=False, border_width=3):
        """
        :param task_id: Task ID.
        :type task_id: int
        :param api: Api instance.
        :type api: :class:`~supervisely.api.api.Api`
        :param v_model: Vue model path.
        :type v_model: str
        :param project_meta: ProjectMeta for annotations.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param col_number: Number of columns.
        :type col_number: int
        :param preview_info: If True, show preview info.
        :type preview_info: bool
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool
        :param resize_on_zoom: If True, resize on zoom.
        :type resize_on_zoom: bool
        :param sync_views: If True, sync views across images.
        :type sync_views: bool
        :param show_preview: If True, show preview panel.
        :type show_preview: bool
        :param selectable: If True, allow selection.
        :type selectable: bool
        :param opacity: Annotation opacity.
        :type opacity: float
        :param show_opacity_header: If True, show opacity control.
        :type show_opacity_header: bool
        :param fill_rectangle: If True, fill rectangles.
        :type fill_rectangle: bool
        :param border_width: Border width for shapes.
        :type border_width: int
        """
        self._task_id = task_id
        self._api = api
        self._v_model = v_model
        self._project_meta = project_meta.clone()
        self._data = {}
        self.col_number = col_number
        self.preview_info = preview_info
        self._need_zoom = False
        self._with_title_url = False
        if not isinstance(self.col_number, int):
            raise ValueError("Columns number must be integer, not {}".format(type(self.col_number).__name__))

        self._options = {
            "enableZoom": enable_zoom,
            "syncViews": sync_views,
            "resizeOnZoom": resize_on_zoom,
            "showPreview": show_preview,
            "selectable": selectable,
            "opacity": opacity,
            "showOpacityInHeader": show_opacity_header,
            "fillRectangle": fill_rectangle,
            "borderWidth": border_width
        }
        self._options_initialized = False

    def add_item(self, title, image_url, ann: Union[Annotation, dict] = None, col_index=None, custom_info: dict = None,
                 zoom_to_figure=None, title_url=None):
        """
        :param title: Image title.
        :type title: str
        :param image_url: Image URL.
        :type image_url: str
        :param ann: Annotation.
        :type ann: :class:`~supervisely.annotation.annotation.Annotation`
        :param col_index: Column index.
        :type col_index: int
        :param custom_info: Custom info.
        :type custom_info: dict
        :param zoom_to_figure: Zoom to figure.
        :type zoom_to_figure: tuple
        :param title_url: Title URL.
        :type title_url: str
        """
        if col_index is not None:
            if col_index <= 0 or col_index > self.col_number:
                raise ValueError("Column number is not correct, check your input data")

        res_ann = Annotation((1, 1))
        if ann is not None:
            if type(ann) is dict:
                res_ann = Annotation.from_json(ann, self._project_meta)
            else:
                res_ann = ann.clone()

        self._data[title] = {"image_url": image_url, "ann": res_ann, "col_index": col_index}

        if zoom_to_figure is not None:
            self._data[title]["zoom_to_figure"] = zoom_to_figure
            self._need_zoom = True

        if title_url is not None:
            self.preview_info = True
            self._with_title_url = True
            self._data[title]["labelingUrl"] = title_url

        if self.preview_info:
            if custom_info is not None:
                self._data[title]["info"] = custom_info
            else:
                self._data[title]["info"] = None

    def add_item_by_id(self, image_id, with_ann=True, col_index=None, info_dict=None,
                 zoom_to_figure=None, title_url=None):
        image_info = self._api.image.get_info_by_id(image_id)
        if with_ann:
            ann_info = self._api.annotation.download(image_id)
            ann = sly.Annotation.from_json(ann_info.annotation, self._project_meta)
        else:
            ann = None

        self.add_item(image_info.name, image_info.full_storage_url, ann, col_index, info_dict, zoom_to_figure, title_url)

    def _get_item_annotation(self, name):
        if self.preview_info:
            if self._with_title_url:
                return {
                    "url": self._data[name]["image_url"],
                    "figures": [label.to_json() for label in self._data[name]["ann"].labels],
                    "title": name,
                    "info": self._data[name]["info"],
                    "labelingUrl": self._data[name]["labelingUrl"]
                }
            else:
                return {
                    "url": self._data[name]["image_url"],
                    "figures": [label.to_json() for label in self._data[name]["ann"].labels],
                    "title": name,
                    "info": self._data[name]["info"]
                }
        else:
            return {
                "url": self._data[name]["image_url"],
                "figures": [label.to_json() for label in self._data[name]["ann"].labels],
                "title": name,
            }

    def update(self, options=True):
        if len(self._data) == 0:
            raise ValueError("Items list is empty")

        gallery_json = self.to_json()

        if options is True or self._options_initialized is False:
            if self._need_zoom:
                self._options["resizeOnZoom"] = True
            self._api.task.set_field(self._task_id, self._v_model, gallery_json)
            self._options_initialized = True
        else:
            self._api.task.set_field(self._task_id, f"{self._v_model}.content", gallery_json["content"])

    def _zoom_to_figure(self, annotations):
        items = self._data.items()
        zoom_to_figure_name = "zoomToFigure"
        for item in items:
            curr_image_name = item[0]
            curr_image_data = item[1]

            if type(curr_image_data["zoom_to_figure"]) is not tuple:
                raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))
            elif type(curr_image_data["zoom_to_figure"]) is None:
                raise ValueError("Option zoom_to_figure not set for {} image".format(curr_image_name))

            zoom_params = {
                "figureId": curr_image_data["zoom_to_figure"][0],
                "factor": curr_image_data["zoom_to_figure"][1]
            }
            annotations[curr_image_name][zoom_to_figure_name] = zoom_params

    def _add_info(self, annotations):
        items = self._data.items()
        for item in items:
            curr_image_name = item[0]
            curr_image_data = item[1]

            annotations[curr_image_name]["info"] = curr_image_data["info"]

    def to_json(self):
        annotations = {}
        layout = [[] for _ in range(self.col_number)]
        index_in_layout = 0

        for curr_image_name, curr_image_data in self._data.items():
            annotations[curr_image_name] = self._get_item_annotation(curr_image_name)

            curr_col_index = curr_image_data["col_index"]
            if curr_col_index is not None:
                layout[curr_col_index - 1].append(curr_image_name)
            else:
                if index_in_layout == self.col_number:
                    index_in_layout = 0
                layout[index_in_layout].append(curr_image_name)
                index_in_layout += 1

        if self._need_zoom:
            self._zoom_to_figure(annotations)

        if self.preview_info:
            self._add_info(annotations)

        return {
            "content": {
                "projectMeta": self._project_meta.to_json(),
                "layout": layout,
                "annotations": annotations
            },
            "options": {
                **self._options
            }
        }



import os
from typing import Dict, List, Optional, Set, Tuple
from uuid import UUID

import magic
import numpy as np

from supervisely import (
    Api,
    PointcloudAnnotation,
    batched,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.api.module_api import ApiField
from supervisely.convert.base_converter import BaseConverter
from supervisely.io.fs import get_file_ext, get_file_name_with_ext
from supervisely.io.json import load_json_file
from supervisely.pointcloud.pointcloud import ALLOWED_POINTCLOUD_EXTENSIONS
from supervisely.pointcloud.pointcloud import validate_ext as validate_pcd_ext
from supervisely.pointcloud_annotation.constants import OBJECT_KEY
from supervisely.video_annotation.key_id_map import KeyIdMap

CONVERTIBLE_EXTENSIONS = [".bin"]


class PointcloudConverter(BaseConverter):
    """Base converter for pointcloud projects (uploads pointclouds, annotations, and optional related images)."""

    allowed_exts = ALLOWED_POINTCLOUD_EXTENSIONS + CONVERTIBLE_EXTENSIONS
    modality = "pointclouds"

    class Item(BaseConverter.BaseItem):
        """Base pointcloud item: path + optional annotation data and related images bundle."""

        def __init__(
            self,
            item_path,
            ann_data=None,
            related_images: Optional[list] = None,
            custom_data: Optional[dict] = None,
        ):
            """
            :param item_path: Path to pointcloud file.
            :type item_path: str
            :param ann_data: Annotation path or data.
            :type ann_data: str, optional
            :param related_images: List of related image tuples.
            :type related_images: list, optional
            :param custom_data: Extra per-item data.
            :type custom_data: dict, optional
            """
            self._name: str = None
            self._path = item_path
            self._ann_data = ann_data
            self._type = "point_cloud"
            self._related_images = related_images if related_images is not None else []
            self._custom_data = custom_data if custom_data is not None else {}

        def create_empty_annotation(self) -> PointcloudAnnotation:
            return PointcloudAnnotation()

        def set_related_images(self, related_images: Tuple[str, str, Optional[str]]) -> None:
            """Adds related image to the item.

            related_images tuple:
                - path to image
                - path to .json with image metadata
                - path to .figures.json (can be None if no figures)
            """
            self._related_images.append(related_images)

    @property
    def format(self):
        return self._converter.format

    @property
    def ann_ext(self):
        return None

    @property
    def key_file_ext(self):
        return None

    @staticmethod
    def validate_ann_file(ann_path, meta=None):
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        existing_pcd_infos = api.pointcloud.get_list(dataset_id)
        existing_pointcloud_names = set([pcd.name for pcd in existing_pcd_infos])
        used_related_image_names: Set[str] = set()
        try:
            existing_pcd_ids = [pcd.id for pcd in existing_pcd_infos]
            for pcd_ids_batch in batched(existing_pcd_ids, batch_size=200):
                related_images = api.pointcloud.get_list_related_images_batch(
                    dataset_id, pcd_ids_batch
                )
                for related_image in related_images:
                    related_image_name = related_image.get(ApiField.NAME)
                    if related_image_name is not None:
                        used_related_image_names.add(related_image_name)
        except Exception as e:
            logger.debug(
                f"Failed to fetch existing related image names for dataset ID:{dataset_id}: {repr(e)}"
            )

        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Uploading pointclouds...")
        else:
            progress_cb = None

        upload_fn = api.pointcloud.upload_paths
        if self.upload_as_links and self.supports_links:
            upload_fn = api.pointcloud.upload_links

        key_id_map = KeyIdMap()
        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []
            anns = []
            for item in batch:
                item.name = generate_free_name(
                    existing_pointcloud_names, item.name, with_ext=True, extend_used_names=True
                )
                item_names.append(item.name)
                if self.upload_as_links:
                    item_paths.append(
                        self.remote_files_map.get(os.path.abspath(item.path), item.path)
                    )
                else:
                    item_paths.append(item.path)

                ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
                anns.append(ann)

            pcd_infos = upload_fn(
                dataset_id,
                item_names,
                item_paths,
            )
            pcd_ids = [pcd_info.id for pcd_info in pcd_infos]
            pcl_to_rimg_figures: Dict[int, Dict[str, List[Dict]]] = {}
            pcl_to_rimg_to_id: Dict[int, Dict[str, int]] = {}
            for pcd_id, ann, item in zip(pcd_ids, anns, batch):
                if ann is not None:
                    api.pointcloud.annotation.append(pcd_id, ann, key_id_map)

                rimg_infos = []
                camera_names = []
                for img_ind, rel_tuple in enumerate(item._related_images):
                    img_path = rel_tuple[0]
                    rimg_ann_path = rel_tuple[1]
                    fig_path = rel_tuple[2] if len(rel_tuple) > 2 else None
                    meta_json = load_json_file(rimg_ann_path)
                    try:
                        if ApiField.META not in meta_json:
                            raise ValueError("Related image meta not found in json file.")
                        if ApiField.NAME not in meta_json:
                            raise ValueError("Related image name not found in json file.")

                        img_hash = None
                        if not self.upload_as_links:
                            img_hash = api.pointcloud.upload_related_image(img_path)

                        if "deviceId" not in meta_json[ApiField.META].keys():
                            camera_names.append(f"CAM_{str(img_ind).zfill(2)}")
                        else:
                            camera_names.append(meta_json[ApiField.META]["deviceId"])

                        related_image_name = generate_free_name(
                            used_related_image_names,
                            meta_json[ApiField.NAME],
                            with_ext=True,
                            extend_used_names=True,
                        )

                        rimage_dict = {
                            ApiField.ENTITY_ID: pcd_id,
                            ApiField.NAME: related_image_name,
                            ApiField.META: meta_json[ApiField.META],
                        }
                        link = None
                        if img_hash is not None:
                            rimage_dict[ApiField.HASH] = img_hash
                        else:
                            link = self.remote_files_map.get(os.path.abspath(img_path), img_path)
                            rimage_dict[ApiField.LINK] = link
                        img_data = img_hash or link
                        rimg_infos.append(rimage_dict)

                        if fig_path is not None and os.path.isfile(fig_path):
                            try:
                                figs_json = load_json_file(fig_path)
                                pcl_to_rimg_figures.setdefault(pcd_id, {})[img_data] = figs_json
                            except Exception as e:
                                logger.debug(f"Failed to read figures json '{fig_path}': {repr(e)}")

                    except Exception as e:
                        logger.warning(
                            f"Failed to upload related image or add it to pointcloud: {repr(e)}"
                        )
                        continue

                # add images for this point cloud
                if len(rimg_infos) > 0:
                    try:
                        uploaded_rimgs = api.pointcloud.add_related_images(rimg_infos, camera_names)
                        # build mapping hash->id
                        for info, uploaded in zip(rimg_infos, uploaded_rimgs):
                            img_data = info.get(ApiField.HASH, info.get(ApiField.LINK, None))
                            img_id = (
                                uploaded.get(ApiField.ID)
                                if isinstance(uploaded, dict)
                                else getattr(uploaded, "id", None)
                            )
                            if img_data is not None and img_id is not None:
                                pcl_to_rimg_to_id.setdefault(pcd_id, {})[img_data] = img_id
                    except Exception as e:
                        logger.debug(f"Failed to add related images to pointcloud: {repr(e)}")

            # ---- upload figures for processed batch ----
            if len(pcl_to_rimg_figures) > 0:
                try:
                    dataset_info = api.dataset.get_info_by_id(dataset_id)
                    project_id = dataset_info.project_id

                    figures_payload: List[Dict] = []

                    for pcl_id, rimg_data_to_figs in pcl_to_rimg_figures.items():
                        rimg_data_to_id = pcl_to_rimg_to_id.get(pcl_id, {})
                        if len(rimg_data_to_id) == 0:
                            continue

                        for img_data, figs_json in rimg_data_to_figs.items():
                            if img_data not in rimg_data_to_id:
                                continue
                            rimg_id = rimg_data_to_id[img_data]
                            for fig in figs_json:
                                try:
                                    fig[ApiField.ENTITY_ID] = rimg_id
                                    fig[ApiField.DATASET_ID] = dataset_id
                                    fig[ApiField.PROJECT_ID] = project_id
                                    if OBJECT_KEY in fig:
                                        fig[ApiField.OBJECT_ID] = key_id_map.get_object_id(
                                            UUID(fig[OBJECT_KEY])
                                        )
                                except Exception as e:
                                    logger.debug(
                                        f"Failed to process figure json for current image. Hash / link = {img_data}: {repr(e)}"
                                    )
                                    continue

                            figures_payload.extend(figs_json)

                        if len(figures_payload) > 0:
                            try:
                                api.image.figure.create_bulk(
                                    figures_json=figures_payload, dataset_id=dataset_id
                                )
                            except Exception as e:
                                logger.debug(
                                    f"Failed to upload figures for related images: {repr(e)}"
                                )
                except Exception as e:
                    logger.debug(f"Unexpected error during related image figures upload: {repr(e)}")

            if log_progress:
                progress_cb(len(batch))

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")

    def _collect_items_if_format_not_detected(self) -> Tuple[List[Item], bool, Set[str]]:
        only_modality_items = True
        unsupported_exts = set()
        pcd_list, rimg_dict, rimg_ann_dict, rimg_fig_dict = [], {}, {}, {}
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                if file in ["key_id_map.json", "meta.json"]:
                    continue

                ext = get_file_ext(full_path)
                if file.endswith(".figures.json"):
                    rimg_fig_dict[file] = full_path
                elif ext == ".json":
                    dir_name = os.path.basename(root)
                    parent_dir_name = os.path.basename(os.path.dirname(root))
                    if any(
                        p.replace("_", " ") in ["images", "related images", "photo context"]
                        for p in [dir_name, parent_dir_name]
                    ) or dir_name.endswith("_pcd"):
                        rimg_ann_dict[file] = full_path
                elif self._is_image_file(full_path):
                    dir_name = os.path.basename(root)
                    if dir_name not in rimg_dict:
                        rimg_dict[dir_name] = []
                    rimg_dict[dir_name].append(full_path)
                elif ext.lower() in self.allowed_exts:
                    try:
                        path = self._convert_to_pcd_if_needed(full_path, ext)
                        if not path:
                            continue
                        pcd_list.append(path)
                    except:
                        pass
                else:
                    only_modality_items = False
                    unsupported_exts.add(ext)

        # create Items
        items = []
        for pcd_path in pcd_list:
            item = self.Item(pcd_path)
            rimg_dir_name = item.name.replace(".pcd", "_pcd")
            rimgs = rimg_dict.get(rimg_dir_name, [])
            for rimg_path in rimgs:
                rimg_ann_name = f"{get_file_name_with_ext(rimg_path)}.json"
                if rimg_ann_name in rimg_ann_dict:
                    rimg_ann_path = rimg_ann_dict[rimg_ann_name]
                    rimg_fig_name = f"{get_file_name_with_ext(rimg_path)}.figures.json"
                    rimg_fig_path = rimg_fig_dict.get(rimg_fig_name, None)
                    if rimg_fig_path is not None and not os.path.exists(rimg_fig_path):
                        rimg_fig_path = None
                    item.set_related_images((rimg_path, rimg_ann_path, rimg_fig_path))
            items.append(item)
        return items, only_modality_items, unsupported_exts

    def _convert_to_pcd_if_needed(self, pcd_path: str, ext: str) -> Optional[str]:
        """Convert point cloud to .pcd format if it is in another supported format."""
        if ext == ".pcd":
            return pcd_path
        elif ext == ".bin":
            if self.upload_as_links:
                logger.warning(
                    f"Uploading .bin pointcloud files as links is not supported. "
                    f"Please convert .bin files to .pcd format before uploading. "
                    f"Skipping conversion for: {pcd_path}"
                )
                return None
            if not self._validate_bin_pointcloud(pcd_path):
                logger.warning(
                    f"The .bin pointcloud file '{pcd_path}' is not valid. Skipping conversion to .pcd."
                )
                return None
            pcd_output_path = pcd_path.replace(".bin", ".pcd")
            self._convert_bin_to_pcd(pcd_path, pcd_output_path)
            return pcd_output_path

    def _validate_bin_pointcloud(self, bin_file: str) -> bool:
        try:
            data = np.fromfile(bin_file, dtype=np.float32)
            if data.size == 0:
                return False

            if data.size % 5 == 0:
                points = data.reshape(-1, 5)
                return bool(np.isfinite(points).all())

            if data.size % 4 == 0:
                points = data.reshape(-1, 4)
                return bool(np.isfinite(points).all())

            return False
        except Exception:
            return False

    def _convert_bin_to_pcd(self, bin_file: str, pcd_file: str) -> None:
        """Convert .bin point cloud file to .pcd format using open3d library (supports 4 or 5 floats per point)."""
        try:
            import open3d as o3d
        except ImportError:
            raise ImportError(
                "open3d is not installed. Cannot convert .bin point cloud to .pcd format. "
                "Please install open3d package to enable this feature."
            )

        try:
            raw = np.fromfile(bin_file, dtype=np.float32)
            if raw.size % 5 == 0:
                arr = raw.reshape(-1, 5)
                points = arr[:, 0:3]
                intensity = arr[:, 3]
            elif raw.size % 4 == 0:
                arr = raw.reshape(-1, 4)
                points = arr[:, 0:3]
                intensity = arr[:, 3]
            else:
                raise ValueError("The number of float32 values is not a multiple of 4 or 5.")
        except ValueError as e:
            raise Exception(
                f"Incorrect data in the BIN pointcloud file: {bin_file}. "
                f"The number of float32 values must be a multiple of 4 (x,y,z,intensity) "
                f"or 5 (x,y,z,intensity,ring_index). Details: {e} "
                "Please ensure that the binary file contains a valid number of elements to be "
                "successfully reshaped into a (N, 4) or (N, 5) array.\n"
            )
        # normalize intensity to [0, 1] for Open3D colors
        intensity = np.nan_to_num(intensity, nan=0.0, posinf=0.0, neginf=0.0)
        if intensity.size > 0:
            i_min = float(np.min(intensity))
            i_max = float(np.max(intensity))
            if i_max > i_min:
                norm_intensity = (intensity - i_min) / (i_max - i_min)
            else:
                norm_intensity = np.zeros_like(intensity)
        else:
            norm_intensity = intensity
        intensity_fake_rgb = np.zeros((norm_intensity.shape[0], 3))
        intensity_fake_rgb[:, 0] = norm_intensity
        pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(points))
        pc.colors = o3d.utility.Vector3dVector(intensity_fake_rgb)
        o3d.io.write_point_cloud(pcd_file, pc)

    @staticmethod
    def _is_image_file(path: str) -> bool:
        try:
            return magic.from_file(path, mime=True).startswith("image/")
        except Exception:
            return False



import copy

import fastapi

from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class ProjectSelector(Widget):
    """Selector for team, workspace, project and optionally datasets; returns selected IDs from state."""

    # @TODO: add Routes project changes events
    # class Routes:
    #     def __init__(self,
    #                  app: fastapi.FastAPI,
    #                  cell_clicked_cb: object = None):
    #         self.app = app
    #         self.routes = {'cell_clicked_cb': cell_clicked_cb}

    def __init__(self,
                 team_id: int = None,
                 workspace_id: int = None,
                 project_id: int = None,
                 team_is_selectable: bool = True,
                 datasets_is_selectable: bool = True,
                 widget_id: str = None):
        """:param team_id: Initial team ID.
        :type team_id: int, optional
        :param workspace_id: Initial workspace ID.
        :type workspace_id: int, optional
        :param project_id: Initial project ID.
        :type project_id: int, optional
        :param team_is_selectable: If True, team can be changed.
        :type team_is_selectable: bool
        :param datasets_is_selectable: If True, datasets can be selected.
        :type datasets_is_selectable: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._team_id = team_id
        self._workspace_id = workspace_id
        self._project_id = project_id

        self._is_selectable = {
            'team': team_is_selectable,
            'datasets': datasets_is_selectable
        }

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            'selectable': copy.copy(self._is_selectable),
            'disabled': False
        }

    def get_json_state(self):
        return {
            'teamId': self._team_id,
            'workspaceId': self._workspace_id,
            'projectId': self._project_id,
            'allDatasets': True,
            'datasetsIds': []
        }

    def get_selected_team_id(self, state):
        return state[self.widget_id]['teamId']

    def get_selected_workspace_id(self, state):
        return state[self.widget_id]['workspaceId']

    def get_selected_project_id(self, state):
        return state[self.widget_id]['projectId']

    def get_selected_datasets(self, state):
        datasets = []
        if self._is_selectable['datasets']:
            datasets = state[self.widget_id]['datasetsIds']
        return datasets

    @property
    def disabled(self):
        return DataJson()[self.widget_id]['disabled']

    @disabled.setter
    def disabled(self, value):
        DataJson()[self.widget_id]['disabled'] = value



from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from typing import Optional, Union


class Image(Widget):
    """Widget that displays an image from URL with optional height/width (px or percent)."""

    def __init__(
        self,
        url: str = "",
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
        widget_id: str = None,
    ):
        """
        :param url: Image URL or data URL.
        :type url: str
        :param height: Height in pixels or CSS value (e.g., "300px", "100%").
        :type height: Union[int, str], optional
        :param width: Width in pixels or CSS value.
        :type width: Union[int, str], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._url = url
        self._height, self._width = self._check_image_size(height=height, width=width)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "url": self._url,
            "height": self._height,
            "width": self._width,
        }

    def get_json_state(self):
        return None

    def set(
        self,
        url: str,
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
    ):
        self._update(url=url, height=height, width=width)

    def clean_up(self):
        self._update(url="", height=None, width=None)

    def set_image_size(
        self,
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
    ):
        self._update(url=self._url, height=height, width=width)

    def _update(
        self,
        url: str = "",
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
    ):
        self._url = url
        self._height, self._width = self._check_image_size(height=height, width=width)
        DataJson()[self.widget_id]["url"] = self._url
        DataJson()[self.widget_id]["height"] = self._height
        DataJson()[self.widget_id]["width"] = self._width
        DataJson().send_changes()

    def _check_image_size(
        self,
        height: Optional[Union[int, str]],
        width: Optional[Union[int, str]],
    ):
        if height is None and width is None:
            return "auto", "100%"

        def _check_single_size(size: Optional[Union[int, str]]):
            if size is None:
                return "auto"
            elif isinstance(size, int):
                return f"{size}px"
            elif isinstance(size, str):
                if size.endswith("px") or size.endswith("%") or size == "auto":
                    return size
                else:
                    raise ValueError(f"size must be in pixels or percent, got '{size}'")
            else:
                raise ValueError(f"size must be int or str, got '{type(size)}'")

        height = _check_single_size(size=height)
        width = _check_single_size(size=width)
        return height, width



from __future__ import annotations

from typing import Dict

from supervisely.api.image_api import ImageApi, ImageInfo
from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import MarkdownWidget, TableWidget


class ModelPredictions(DetectionVisMetric):
    """Per-image prediction counts table for detection reports."""

    MARKDOWN = "model_predictions"
    TABLE = "model_predictions"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True
        self._row_ids = None  #  TODO: check if this is used

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_predictions_table
        return MarkdownWidget(self.MARKDOWN, "Prediction details for every image", text)

    @property
    def table(self) -> TableWidget:
        tmp = set([d.pred_image_info.name for d in self.eval_result.matched_pair_data.values()])
        df = self.eval_result.mp.prediction_table().round(2)
        df = df[df["Image name"].isin(tmp)]
        columns_options = [
            {"maxWidth": "225px"},
            {"subtitle": "objects count", "tooltip": "Number of ground truth objects on the image"},
            {"subtitle": "objects count", "tooltip": "Number of predicted objects on the image"},
            {
                "subtitle": "objects count",
                "tooltip": "True Positive objects count (correctly detected)",
            },
            {
                "subtitle": "objects count",
                "tooltip": "False Positive objects count (incorrect detections)",
            },
            {
                "subtitle": "objects count",
                "tooltip": "False Negative objects count (missed detections)",
            },
            {"maxValue": 1, "tooltip": "Precision (positive predictive value)"},
            {"maxValue": 1, "tooltip": "Recall (sensitivity)"},
            {"maxValue": 1, "tooltip": "F1 score (harmonic mean of precision and recall)"},
        ]

        columns = df.columns.tolist()[1:]  # exclude sly_id
        content = []

        key_mapping = {}
        for old, new in zip(ImageInfo._fields, ImageApi.info_sequence()):
            key_mapping[old] = new

        self._row_ids = []
        df = df.replace({float("nan"): None}) # replace NaN / float("nan") with None

        for row in df.values.tolist():
            sly_id = row.pop(0)
            info = self.eval_result.matched_pair_data[sly_id].gt_image_info

            dct = {
                "row": {key_mapping[k]: v for k, v in info._asdict().items()},
                "id": info.name,
                "items": row,
            }

            self._row_ids.append(dct["id"])
            content.append(dct)

        data = {
            "columns": columns,
            "columnsOptions": columns_options,
            "content": content,
        }
        table = TableWidget(
            name=self.TABLE,
            data=data,
            fix_columns=1,
        )
        table.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
        )
        return table

    def get_click_data(self) -> Dict:
        res = {}
        res["layoutTemplate"] = [
            {"skipObjectTagsFiltering": True, "columnTitle": "Ground Truth"},
            {"skipObjectTagsFiltering": ["outcome"], "columnTitle": "Prediction"},
            {"skipObjectTagsFiltering": ["confidence"], "columnTitle": "Difference"},
        ]
        click_data = res.setdefault("clickData", {})

        default_filters = [
            {
                "type": "tag",
                "tagId": "confidence",
                "value": [self.eval_result.mp.conf_threshold, 1],
            },
            # {"type": "tag", "tagId": "outcome", "value": "FP"},
        ]

        for pairs_data in self.eval_result.matched_pair_data.values():
            gt = pairs_data.gt_image_info
            pred = pairs_data.pred_image_info
            diff = pairs_data.diff_image_info
            assert gt.name == pred.name == diff.name
            key = click_data.setdefault(str(pred.name), {})
            key["imagesIds"] = [gt.id, pred.id, diff.id]
            key["filters"] = default_filters
            key["title"] = f"Image: {pred.name}"
            image_id = pred.id
            ann_json = pairs_data.pred_annotation.to_json()
            assert image_id == pred.id
            object_bindings = []
            for obj in ann_json["objects"]:
                for tag in obj["tags"]:
                    if tag["name"] == "matched_gt_id":
                        object_bindings.append(
                            [
                                {
                                    "id": obj["id"],
                                    "annotationKey": image_id,
                                },
                                {
                                    "id": int(tag["value"]),
                                    "annotationKey": gt.id,
                                },
                            ]
                        )

            image_id = diff.id
            ann_json = pairs_data.diff_annotation.to_json()
            assert image_id == diff.id
            for obj in ann_json["objects"]:
                for tag in obj["tags"]:
                    if tag["name"] == "matched_gt_id":
                        object_bindings.append(
                            [
                                {
                                    "id": obj["id"],
                                    "annotationKey": image_id,
                                },
                                {
                                    "id": int(tag["value"]),
                                    "annotationKey": pred.id,
                                },
                            ]
                        )
            key["objectsBindings"] = object_bindings

        return res



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class Container(Widget):
    """Layout container for arranging child widgets vertically or horizontally with gap and fractions."""

    def __init__(
        self,
        widgets: List[Widget] = [],
        direction: Literal["vertical", "horizontal"] = "vertical",
        gap: int = 10,
        fractions: List[int] = None,
        widget_id: str = None,
    ):
        """
        :param widgets: List of child widgets.
        :type widgets: List[Widget]
        :param direction: "vertical" or "horizontal".
        :type direction: Literal["vertical", "horizontal"]
        :param gap: Gap between widgets in pixels.
        :type gap: int
        :param fractions: Width fractions per widget (horizontal only).
        :type fractions: List[int], optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional

        :raises ValueError: If fractions used with vertical or len mismatch.
        """
        self._widgets = widgets
        self._direction = direction
        self._gap = gap
        if self._direction == "vertical" and fractions is not None:
            raise ValueError("fractions can be defined only with horizontal direction")

        if fractions is not None and len(widgets) != len(fractions):
            raise ValueError(
                "len(widgets) != len(fractions): fractions have to be defined for all widgets"
            )
        self._fractions = fractions
        self._flex_direction = "column"
        if direction == "horizontal":
            self._flex_direction = "row"
            if self._fractions is None:
                self._fractions = [1] * len(self._widgets)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return None

    def get_json_state(self):
        return None



# coding: utf-8
from copy import copy
from typing import List, Optional

MATCH_ALL = '__all__'


def is_name_included(name: str, enabled_names: List[str]) -> bool:
    """"""
    return (enabled_names == MATCH_ALL) or (type(enabled_names) != str and name in enabled_names)


class Renamer:
    """
    Rule-based name transformer used by mappers when cloning object classes.
    """

    ADD_SUFFIX = 'add_suffix'
    """"""
    SAVE_CLASSES = 'save_classes'  # Deprecated. Use SAVE_NAMES in new code.
    """"""
    SAVE_NAMES = 'save_names'  # New field with more generic name.
    """"""

    def __init__(self, add_suffix: str = "", save_names: Optional[List[str]] = None):
        """
        :param add_suffix: Suffix to be added to the processed object names.
        :type add_suffix: str
        :param save_names: List of the object names to which you will need to add a suffix.
        :type save_names: list
        """
        self._add_suffix = add_suffix
        self._save_names = copy(save_names) if save_names is not None else MATCH_ALL

    def rename(self, name):
        """
        The function add special suffix to input name
        :param name: name to be changed
        :returns: new name
        """
        return (name + self._add_suffix) if is_name_included(name, self._save_names) else None

    def to_json(self):
        """
        The function to_json convert Renamer to json format
        :returns: Renamer in json format
        :rtype: dict
        """
        return {Renamer.ADD_SUFFIX: self._add_suffix, Renamer.SAVE_CLASSES: self._save_names}

    @staticmethod
    def from_json(renamer_json):
        """
        The function from_json convert Renamer from json format to Renamer class object.
        :param renamer_json: Renamer in json format
        :returns: Renamer object
        :rtype: :class:`~supervisely.annotation.renamer.Renamer`
        """
        enabled_names = renamer_json.get(Renamer.SAVE_NAMES)
        if enabled_names is None:
            enabled_names = renamer_json.get(Renamer.SAVE_CLASSES)
        add_suffix = renamer_json.get(Renamer.ADD_SUFFIX, '')
        return Renamer(add_suffix=add_suffix, save_names=enabled_names)



import hashlib
import time
from typing import Dict, List, Optional, Tuple, Union

from fastapi import FastAPI, WebSocket, WebSocketDisconnect

import supervisely.io.env as sly_env
from supervisely.app.singleton import Singleton


class WebsocketManager(metaclass=Singleton):
    """Singleton manager for WebSocket connections, broadcasting and user cookie mapping."""

    def __init__(self, path="/sly-app-ws"):
        """
        :param path: WebSocket route path.
        :type path: str
        """
        self.app = None
        self.path = path
        self.active_connections: List[WebSocket] = []
        self._connection_users: Dict[WebSocket, Optional[Union[int, str]]] = {}
        self._cookie_user_map: Dict[str, Tuple[Union[int, str], float]] = {}
        self._cookie_ttl_seconds = 60 * 60

    def set_app(self, app: FastAPI):
        if self.app is not None:
            return
        self.app = app
        self.app.add_api_websocket_route(path=self.path, endpoint=self.endpoint)

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        user_id = self._resolve_user_id(websocket)
        self.active_connections.append(websocket)
        self._connection_users[websocket] = user_id

    def disconnect(self, websocket: WebSocket):
        if websocket in self.active_connections:
            self.active_connections.remove(websocket)
        self._connection_users.pop(websocket, None)

    def remember_user_cookie(
        self, cookie_header: Optional[str], user_id: Optional[Union[int, str]]
    ):
        if cookie_header is None or user_id is None:
            return
        fingerprint = self._cookie_fingerprint(cookie_header)
        if fingerprint is None:
            return
        self._purge_cookie_cache()
        self._cookie_user_map[fingerprint] = (user_id, time.monotonic())

    async def broadcast(self, d: dict, user_id: Optional[Union[int, str]] = None):
        if sly_env.is_multiuser_mode_enabled():
            if user_id is None:
                user_id = sly_env.user_from_multiuser_app()
            if user_id is None:
                targets = list(self.active_connections)
            else:
                targets = [
                    connection
                    for connection in self.active_connections
                    if self._connection_users.get(connection) == user_id
                ]
        else:
            targets = list(self.active_connections)

        for connection in list(targets):
            await connection.send_json(d)

    async def endpoint(self, websocket: WebSocket):
        await self.connect(websocket)
        try:
            while True:
                data = await websocket.receive_text()
        except WebSocketDisconnect:
            self.disconnect(websocket)

    def _resolve_user_id(self, websocket: WebSocket) -> Optional[int]:
        if not sly_env.is_multiuser_mode_enabled():
            return None
        query_user = websocket.query_params.get("userId")
        if query_user is not None:
            try:
                return int(query_user)
            except ValueError:
                pass
        fingerprint = self._cookie_fingerprint(websocket.headers.get("cookie"))
        if fingerprint is None:
            return None
        cached = self._cookie_user_map.get(fingerprint)
        if cached is None:
            return None
        user_id, ts = cached
        if time.monotonic() - ts > self._cookie_ttl_seconds:
            self._cookie_user_map.pop(fingerprint, None)
            return None
        return user_id

    @staticmethod
    def _cookie_fingerprint(cookie_header: Optional[str]) -> Optional[str]:
        if not cookie_header:
            return None
        return hashlib.sha256(cookie_header.encode("utf-8")).hexdigest()

    def _purge_cookie_cache(self) -> None:
        if not self._cookie_user_map:
            return
        cutoff = time.monotonic() - self._cookie_ttl_seconds
        expired = [key for key, (_, ts) in self._cookie_user_map.items() if ts < cutoff]
        for key in expired:
            self._cookie_user_map.pop(key, None)
